# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Routines for configuring Glance
"""

import logging
import logging.config
import logging.handlers
import os

from oslo_config import cfg
from oslo_middleware import cors
from oslo_policy import policy
from paste import deploy

from glance.i18n import _
from glance.version import version_info as version

paste_deploy_opts = [
    cfg.StrOpt('flavor',
               sample_default='keystone',
               help=_("""
Deployment flavor to use in the server application pipeline.

Provide a string value representing the appropriate deployment
flavor used in the server application pipleline. This is typically
the partial name of a pipeline in the paste configuration file with
the service name removed.

For example, if your paste section name in the paste configuration
file is [pipeline:glance-api-keystone], set ``flavor`` to
``keystone``.

Possible values:
    * String value representing a partial pipeline name.

Related Options:
    * config_file

""")),
    cfg.StrOpt('config_file',
               sample_default='glance-api-paste.ini',
               help=_("""
Name of the paste configuration file.

Provide a string value representing the name of the paste
configuration file to use for configuring piplelines for
server application deployments.

NOTES:
    * Provide the name or the path relative to the glance directory
      for the paste configuration file and not the absolute path.
    * The sample paste configuration file shipped with Glance need
      not be edited in most cases as it comes with ready-made
      pipelines for all common deployment flavors.

If no value is specified for this option, the ``paste.ini`` file
with the prefix of the corresponding Glance service's configuration
file name will be searched for in the known configuration
directories. (For example, if this option is missing from or has no
value set in ``glance-api.conf``, the service will look for a file
named ``glance-api-paste.ini``.) If the paste configuration file is
not found, the service will not start.

Possible values:
    * A string value representing the name of the paste configuration
      file.

Related Options:
    * flavor

""")),
]
image_format_opts = [
    cfg.ListOpt('container_formats',
                default=['ami', 'ari', 'aki', 'bare', 'ovf', 'ova', 'docker'],
                help=_("Supported values for the 'container_format' "
                       "image attribute"),
                deprecated_opts=[cfg.DeprecatedOpt('container_formats',
                                                   group='DEFAULT')]),
    cfg.ListOpt('disk_formats',
                default=['ami', 'ari', 'aki', 'vhd', 'vhdx', 'vmdk', 'raw',
                         'qcow2', 'vdi', 'iso'],
                help=_("Supported values for the 'disk_format' "
                       "image attribute"),
                deprecated_opts=[cfg.DeprecatedOpt('disk_formats',
                                                   group='DEFAULT')]),
]
task_opts = [
    cfg.IntOpt('task_time_to_live',
               default=48,
               help=_("Time in hours for which a task lives after, either "
                      "succeeding or failing"),
               deprecated_opts=[cfg.DeprecatedOpt('task_time_to_live',
                                                  group='DEFAULT')]),
    cfg.StrOpt('task_executor',
               default='taskflow',
               help=_("""
Task executor to be used to run task scripts.

Provide a string value representing the executor to use for task
executions. By default, ``TaskFlow`` executor is used.

``TaskFlow`` helps make task executions easy, consistent, scalable
and reliable. It also enables creation of lightweight task objects
and/or functions that are combined together into flows in a
declarative manner.

Possible values:
    * taskflow

Related Options:
    * None

""")),
    cfg.StrOpt('work_dir',
               sample_default='/work_dir',
               help=_("""
Absolute path to the work directory to use for asynchronous
task operations.

The directory set here will be used to operate over images -
normally before they are imported in the destination store.

NOTE: When providing a value for ``work_dir``, please make sure
that enough space is provided for concurrent tasks to run
efficiently without running out of space.

A rough estimation can be done by multiplying the number of
``max_workers`` with an average image size (e.g 500MB). The image
size estimation should be done based on the average size in your
deployment. Note that depending on the tasks running you may need
to multiply this number by some factor depending on what the task
does. For example, you may want to double the available size if
image conversion is enabled. All this being said, remember these
are just estimations and you should do them based on the worst
case scenario and be prepared to act in case they were wrong.

Possible values:
    * String value representing the absolute path to the working
      directory

Related Options:
    * None

""")),
]

_DEPRECATE_GLANCE_V1_MSG = _('The Images (Glance) version 1 API has been '
                             'DEPRECATED in the Newton release and will be '
                             'removed on or after Pike release, following '
                             'the standard OpenStack deprecation policy. '
                             'Hence, the configuration options specific to '
                             'the Images (Glance) v1 API are hereby '
                             'deprecated and subject to removal. Operators '
                             'are advised to deploy the Images (Glance) v2 '
                             'API.')

common_opts = [
    cfg.BoolOpt('allow_additional_image_properties', default=True,
                help=_("""
Allow users to add additional/custom properties to images.

Glance defines a standard set of properties (in its schema) that
appear on every image. These properties are also known as
``base properties``. In addition to these properties, Glance
allows users to add custom properties to images. These are known
as ``additional properties``.

By default, this configuration option is set to ``True`` and users
are allowed to add additional properties. The number of additional
properties that can be added to an image can be controlled via
``image_property_quota`` configuration option.

Possible values:
    * True
    * False

Related options:
    * image_property_quota

""")),
    cfg.IntOpt('image_member_quota', default=128,
               help=_("""
Maximum number of image members per image.

This limits the maximum of users an image can be shared with. Any negative
value is interpreted as unlimited.

Related options:
    * None

""")),
    cfg.IntOpt('image_property_quota', default=128,
               help=_("""
Maximum number of properties allowed on an image.

This enforces an upper limit on the number of additional properties an image
can have. Any negative value is interpreted as unlimited.

NOTE: This won't have any impact if additional properties are disabled. Please
refer to ``allow_additional_image_properties``.

Related options:
    * ``allow_additional_image_properties``

""")),
    cfg.IntOpt('image_tag_quota', default=128,
               help=_("""
Maximum number of tags allowed on an image.

Any negative value is interpreted as unlimited.

Related options:
    * None

""")),
    cfg.IntOpt('image_location_quota', default=10,
               help=_("""
Maximum number of locations allowed on an image.

Any negative value is interpreted as unlimited.

Related options:
    * None

""")),
    # TODO(abashmak): Add choices parameter to this option:
    # choices('glance.db.sqlalchemy.api',
    #         'glance.db.registry.api',
    #         'glance.db.simple.api')
    # This will require a fix to the functional tests which
    # set this option to a test version of the registry api module:
    # (glance.tests.functional.v2.registry_data_api), in order to
    # bypass keystone authentication for the Registry service.
    # All such tests are contained in:
    # glance/tests/functional/v2/test_images.py
    cfg.StrOpt('data_api',
               default='glance.db.sqlalchemy.api',
               help=_("""
Python module path of data access API.

Specifies the path to the API to use for accessing the data model.
This option determines how the image catalog data will be accessed.

Possible values:
    * glance.db.sqlalchemy.api
    * glance.db.registry.api
    * glance.db.simple.api

If this option is set to ``glance.db.sqlalchemy.api`` then the image
catalog data is stored in and read from the database via the
SQLAlchemy Core and ORM APIs.

Setting this option to ``glance.db.registry.api`` will force all
database access requests to be routed through the Registry service.
This avoids data access from the Glance API nodes for an added layer
of security, scalability and manageability.

NOTE: In v2 OpenStack Images API, the registry service is optional.
In order to use the Registry API in v2, the option
``enable_v2_registry`` must be set to ``True``.

Finally, when this configuration option is set to
``glance.db.simple.api``, image catalog data is stored in and read
from an in-memory data structure. This is primarily used for testing.

Related options:
    * enable_v2_api
    * enable_v2_registry

""")),
    cfg.IntOpt('limit_param_default', default=25, min=1,
               help=_("""
The default number of results to return for a request.

Responses to certain API requests, like list images, may return
multiple items. The number of results returned can be explicitly
controlled by specifying the ``limit`` parameter in the API request.
However, if a ``limit`` parameter is not specified, this
configuration value will be used as the default number of results to
be returned for any API request.

NOTES:
    * The value of this configuration option may not be greater than
      the value specified by ``api_limit_max``.
    * Setting this to a very large value may slow down database
      queries and increase response times. Setting this to a
      very low value may result in poor user experience.

Possible values:
    * Any positive integer

Related options:
    * api_limit_max

""")),
    cfg.IntOpt('api_limit_max', default=1000, min=1,
               help=_("""
Maximum number of results that could be returned by a request.

As described in the help text of ``limit_param_default``, some
requests may return multiple results. The number of results to be
returned are governed either by the ``limit`` parameter in the
request or the ``limit_param_default`` configuration option.
The value in either case, can't be greater than the absolute maximum
defined by this configuration option. Anything greater than this
value is trimmed down to the maximum value defined here.

NOTE: Setting this to a very large value may slow down database
      queries and increase response times. Setting this to a
      very low value may result in poor user experience.

Possible values:
    * Any positive integer

Related options:
    * limit_param_default

""")),
    cfg.BoolOpt('show_image_direct_url', default=False,
                help=_("""
Show direct image location when returning an image.

This configuration option indicates whether to show the direct image
location when returning image details to the user. The direct image
location is where the image data is stored in backend storage. This
image location is shown under the image property ``direct_url``.

When multiple image locations exist for an image, the best location
is displayed based on the location strategy indicated by the
configuration option ``location_strategy``.

NOTES:
    * Revealing image locations can present a GRAVE SECURITY RISK as
      image locations can sometimes include credentials. Hence, this
      is set to ``False`` by default. Set this to ``True`` with
      EXTREME CAUTION and ONLY IF you know what you are doing!
    * If an operator wishes to avoid showing any image location(s)
      to the user, then both this option and
      ``show_multiple_locations`` MUST be set to ``False``.

Possible values:
    * True
    * False

Related options:
    * show_multiple_locations
    * location_strategy

""")),
    # NOTE(flaper87): The policy.json file should be updated and the locaiton
    # related rules set to admin only once this option is finally removed.
    cfg.BoolOpt('show_multiple_locations', default=False,
                deprecated_for_removal=True,
                deprecated_reason=_('This option will be removed in the Ocata '
                                    'release because the same functionality '
                                    'can be achieved with greater granularity '
                                    'by using policies. Please see the Newton '
                                    'release notes for more information.'),
                deprecated_since='Newton',
                help=_("""
Show all image locations when returning an image.

This configuration option indicates whether to show all the image
locations when returning image details to the user. When multiple
image locations exist for an image, the locations are ordered based
on the location strategy indicated by the configuration opt
``location_strategy``. The image locations are shown under the
image property ``locations``.

NOTES:
    * Revealing image locations can present a GRAVE SECURITY RISK as
      image locations can sometimes include credentials. Hence, this
      is set to ``False`` by default. Set this to ``True`` with
      EXTREME CAUTION and ONLY IF you know what you are doing!
    * If an operator wishes to avoid showing any image location(s)
      to the user, then both this option and
      ``show_image_direct_url`` MUST be set to ``False``.

Possible values:
    * True
    * False

Related options:
    * show_image_direct_url
    * location_strategy

""")),
    cfg.IntOpt('image_size_cap', default=1099511627776, min=1,
               max=9223372036854775808,
               help=_("""
Maximum size of image a user can upload in bytes.

An image upload greater than the size mentioned here would result
in an image creation failure. This configuration option defaults to
1099511627776 bytes (1 TiB).

NOTES:
    * This value should only be increased after careful
      consideration and must be set less than or equal to
      8 EiB (9223372036854775808).
    * This value must be set with careful consideration of the
      backend storage capacity. Setting this to a very low value
      may result in a large number of image failures. And, setting
      this to a very large value may result in faster consumption
      of storage. Hence, this must be set according to the nature of
      images created and storage capacity available.

Possible values:
    * Any positive number less than or equal to 9223372036854775808

""")),
    cfg.StrOpt('user_storage_quota', default='0',
               help=_("""
Maximum amount of image storage per tenant.

This enforces an upper limit on the cumulative storage consumed by all images
of a tenant across all stores. This is a per-tenant limit.

The default unit for this configuration option is Bytes. However, storage
units can be specified using case-sensitive literals ``B``, ``KB``, ``MB``,
``GB`` and ``TB`` representing Bytes, KiloBytes, MegaBytes, GigaBytes and
TeraBytes respectively. Note that there should not be any space between the
value and unit. Value ``0`` signifies no quota enforcement. Negative values
are invalid and result in errors.

Possible values:
    * A string that is a valid concatenation of a non-negative integer
      representing the storage value and an optional string literal
      representing storage units as mentioned above.

Related options:
    * None

""")),
    # NOTE(nikhil): Even though deprecated, the configuration option
    # ``enable_v1_api`` is set to True by default on purpose. Having it enabled
    # helps the projects that haven't been able to fully move to v2 yet by
    # keeping the devstack setup to use glance v1 as well. We need to switch it
    # to False by default soon after Newton is cut so that we can identify the
    # projects that haven't moved to v2 yet and start having some interesting
    # conversations with them. Switching to False in Newton may result into
    # destabilizing the gate and affect the release.
    cfg.BoolOpt('enable_v1_api',
                default=True,
                deprecated_reason=_DEPRECATE_GLANCE_V1_MSG,
                deprecated_since='Newton',
                help=_("""
Deploy the v1 OpenStack Images API.

When this option is set to ``True``, Glance service will respond to
requests on registered endpoints conforming to the v1 OpenStack
Images API.

NOTES:
    * If this option is enabled, then ``enable_v1_registry`` must
      also be set to ``True`` to enable mandatory usage of Registry
      service with v1 API.

    * If this option is disabled, then the ``enable_v1_registry``
      option, which is enabled by default, is also recommended
      to be disabled.

    * This option is separate from ``enable_v2_api``, both v1 and v2
      OpenStack Images API can be deployed independent of each
      other.

    * If deploying only the v2 Images API, this option, which is
      enabled by default, should be disabled.

Possible values:
    * True
    * False

Related options:
    * enable_v1_registry
    * enable_v2_api

""")),
    cfg.BoolOpt('enable_v2_api',
                default=True,
                deprecated_reason=_('The Images (Glance) version 1 API has '
                                    'been DEPRECATED in the Newton release. '
                                    'It will be removed on or after Pike '
                                    'release, following the standard '
                                    'OpenStack deprecation policy. Once we '
                                    'remove the Images (Glance) v1 API, only '
                                    'the Images (Glance) v2 API can be '
                                    'deployed and will be enabled by default '
                                    'making this option redundant.'),
                deprecated_since='Newton',
                help=_("""
Deploy the v2 OpenStack Images API.

When this option is set to ``True``, Glance service will respond
to requests on registered endpoints conforming to the v2 OpenStack
Images API.

NOTES:
    * If this option is disabled, then the ``enable_v2_registry``
      option, which is enabled by default, is also recommended
      to be disabled.

    * This option is separate from ``enable_v1_api``, both v1 and v2
      OpenStack Images API can be deployed independent of each
      other.

    * If deploying only the v1 Images API, this option, which is
      enabled by default, should be disabled.

Possible values:
    * True
    * False

Related options:
    * enable_v2_registry
    * enable_v1_api

""")),
    cfg.BoolOpt('enable_v1_registry',
                default=True,
                deprecated_reason=_DEPRECATE_GLANCE_V1_MSG,
                deprecated_since='Newton',
                help=_("""
Deploy the v1 API Registry service.

When this option is set to ``True``, the Registry service
will be enabled in Glance for v1 API requests.

NOTES:
    * Use of Registry is mandatory in v1 API, so this option must
      be set to ``True`` if the ``enable_v1_api`` option is enabled.

    * If deploying only the v2 OpenStack Images API, this option,
      which is enabled by default, should be disabled.

Possible values:
    * True
    * False

Related options:
    * enable_v1_api

""")),
    cfg.BoolOpt('enable_v2_registry',
                default=True,
                help=_("""
Deploy the v2 API Registry service.

When this option is set to ``True``, the Registry service
will be enabled in Glance for v2 API requests.

NOTES:
    * Use of Registry is optional in v2 API, so this option
      must only be enabled if both ``enable_v2_api`` is set to
      ``True`` and the ``data_api`` option is set to
      ``glance.db.registry.api``.

    * If deploying only the v1 OpenStack Images API, this option,
      which is enabled by default, should be disabled.

Possible values:
    * True
    * False

Related options:
    * enable_v2_api
    * data_api

""")),
    cfg.StrOpt('pydev_worker_debug_host',
               sample_default='localhost',
               help=_("""
Host address of the pydev server.

Provide a string value representing the hostname or IP of the
pydev server to use for debugging. The pydev server listens for
debug connections on this address, facilitating remote debugging
in Glance.

Possible values:
    * Valid hostname
    * Valid IP address

Related options:
    * None

""")),
    cfg.PortOpt('pydev_worker_debug_port',
                default=5678,
                help=_("""
Port number that the pydev server will listen on.

Provide a port number to bind the pydev server to. The pydev
process accepts debug connections on this port and facilitates
remote debugging in Glance.

Possible values:
    * A valid port number

Related options:
    * None

""")),
    cfg.StrOpt('metadata_encryption_key',
               secret=True,
               help=_("""
AES key for encrypting store location metadata.

Provide a string value representing the AES cipher to use for
encrypting Glance store metadata.

NOTE: The AES key to use must be set to a random string of length
16, 24 or 32 bytes.

Possible values:
    * String value representing a valid AES key

Related options:
    * None

""")),
    cfg.StrOpt('digest_algorithm',
               default='sha256',
               help=_("""
Digest algorithm to use for digital signature.

Provide a string value representing the digest algorithm to
use for generating digital signatures. By default, ``sha256``
is used.

To get a list of the available algorithms supported by the version
of OpenSSL on your platform, run the command:
``openssl list-message-digest-algorithms``.
Examples are 'sha1', 'sha256', and 'sha512'.

NOTE: ``digest_algorithm`` is not related to Glance's image signing
and verification. It is only used to sign the universally unique
identifier (UUID) as a part of the certificate file and key file
validation.

Possible values:
    * An OpenSSL message digest algorithm identifier

Relation options:
    * None

""")),
]

CONF = cfg.CONF
CONF.register_opts(paste_deploy_opts, group='paste_deploy')
CONF.register_opts(image_format_opts, group='image_format')
CONF.register_opts(task_opts, group='task')
CONF.register_opts(common_opts)
policy.Enforcer(CONF)


def parse_args(args=None, usage=None, default_config_files=None):
    CONF(args=args,
         project='glance',
         version=version.cached_version_string(),
         usage=usage,
         default_config_files=default_config_files)


def parse_cache_args(args=None):
    config_files = cfg.find_config_files(project='glance', prog='glance-cache')
    parse_args(args=args, default_config_files=config_files)


def _get_deployment_flavor(flavor=None):
    """
    Retrieve the paste_deploy.flavor config item, formatted appropriately
    for appending to the application name.

    :param flavor: if specified, use this setting rather than the
                   paste_deploy.flavor configuration setting
    """
    if not flavor:
        flavor = CONF.paste_deploy.flavor
    return '' if not flavor else ('-' + flavor)


def _get_paste_config_path():
    paste_suffix = '-paste.ini'
    conf_suffix = '.conf'
    if CONF.config_file:
        # Assume paste config is in a paste.ini file corresponding
        # to the last config file
        path = CONF.config_file[-1].replace(conf_suffix, paste_suffix)
    else:
        path = CONF.prog + paste_suffix
    return CONF.find_file(os.path.basename(path))


def _get_deployment_config_file():
    """
    Retrieve the deployment_config_file config item, formatted as an
    absolute pathname.
    """
    path = CONF.paste_deploy.config_file
    if not path:
        path = _get_paste_config_path()
    if not path:
        msg = _("Unable to locate paste config file for %s.") % CONF.prog
        raise RuntimeError(msg)
    return os.path.abspath(path)


def load_paste_app(app_name, flavor=None, conf_file=None):
    """
    Builds and returns a WSGI app from a paste config file.

    We assume the last config file specified in the supplied ConfigOpts
    object is the paste config file, if conf_file is None.

    :param app_name: name of the application to load
    :param flavor: name of the variant of the application to load
    :param conf_file: path to the paste config file

    :raises: RuntimeError when config file cannot be located or application
            cannot be loaded from config file
    """
    # append the deployment flavor to the application name,
    # in order to identify the appropriate paste pipeline
    app_name += _get_deployment_flavor(flavor)

    if not conf_file:
        conf_file = _get_deployment_config_file()

    try:
        logger = logging.getLogger(__name__)
        logger.debug("Loading %(app_name)s from %(conf_file)s",
                     {'conf_file': conf_file, 'app_name': app_name})

        app = deploy.loadapp("config:%s" % conf_file, name=app_name)

        # Log the options used when starting if we're in debug mode...
        if CONF.debug:
            CONF.log_opt_values(logger, logging.DEBUG)

        return app
    except (LookupError, ImportError) as e:
        msg = (_("Unable to load %(app_name)s from "
                 "configuration file %(conf_file)s."
                 "\nGot: %(e)r") % {'app_name': app_name,
                                    'conf_file': conf_file,
                                    'e': e})
        logger.error(msg)
        raise RuntimeError(msg)


def set_config_defaults():
    """This method updates all configuration default values."""
    set_cors_middleware_defaults()


def set_cors_middleware_defaults():
    """Update default configuration options for oslo.middleware."""
    # CORS Defaults
    # TODO(krotscheck): Update with https://review.openstack.org/#/c/285368/
    cfg.set_defaults(cors.CORS_OPTS,
                     allow_headers=['Content-MD5',
                                    'X-Image-Meta-Checksum',
                                    'X-Storage-Token',
                                    'Accept-Encoding',
                                    'X-Auth-Token',
                                    'X-Identity-Status',
                                    'X-Roles',
                                    'X-Service-Catalog',
                                    'X-User-Id',
                                    'X-Tenant-Id',
                                    'X-OpenStack-Request-ID'],
                     expose_headers=['X-Image-Meta-Checksum',
                                     'X-Auth-Token',
                                     'X-Subject-Token',
                                     'X-Service-Token',
                                     'X-OpenStack-Request-ID'],
                     allow_methods=['GET',
                                    'PUT',
                                    'POST',
                                    'DELETE',
                                    'PATCH']
                     )



# Copyright 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from contextlib import contextmanager

import glance_store
import mock
from mock import patch
import webob.exc

from glance.api.v1 import upload_utils
from glance.common import exception
from glance.common import store_utils
from glance.common import utils
import glance.registry.client.v1.api as registry
from glance.tests.unit import base
import glance.tests.unit.utils as unit_test_utils


class TestUploadUtils(base.StoreClearingUnitTest):
    def setUp(self):
        super(TestUploadUtils, self).setUp()
        self.config(debug=True)

    def tearDown(self):
        super(TestUploadUtils, self).tearDown()

    def test_initiate_delete(self):
        req = unit_test_utils.get_fake_request()
        location = {"url": "file://foo/bar",
                    "metadata": {},
                    "status": "active"}
        id = unit_test_utils.UUID1

        with patch.object(store_utils,
                          "safe_delete_from_backend") as mock_store_utils:
            upload_utils.initiate_deletion(req, location, id)
            mock_store_utils.assert_called_once_with(req.context,
                                                     id,
                                                     location)

    def test_initiate_delete_with_delayed_delete(self):
        self.config(delayed_delete=True)
        req = unit_test_utils.get_fake_request()
        location = {"url": "file://foo/bar",
                    "metadata": {},
                    "status": "active"}
        id = unit_test_utils.UUID1

        with patch.object(store_utils, "schedule_delayed_delete_from_backend",
                          return_value=True) as mock_store_utils:
            upload_utils.initiate_deletion(req, location, id)
            mock_store_utils.assert_called_once_with(req.context,
                                                     id,
                                                     location)

    def test_safe_kill(self):
        req = unit_test_utils.get_fake_request()
        id = unit_test_utils.UUID1

        with patch.object(registry, "update_image_metadata") as mock_registry:
            upload_utils.safe_kill(req, id, 'saving')
            mock_registry.assert_called_once_with(req.context, id,
                                                  {'status': 'killed'},
                                                  from_state='saving')

    def test_safe_kill_with_error(self):
        req = unit_test_utils.get_fake_request()
        id = unit_test_utils.UUID1

        with patch.object(registry, "update_image_metadata",
                          side_effect=Exception()) as mock_registry:
            upload_utils.safe_kill(req, id, 'saving')
            mock_registry.assert_called_once_with(req.context, id,
                                                  {'status': 'killed'},
                                                  from_state='saving')

    @contextmanager
    def _get_store_and_notifier(self, image_size=10, ext_update_data=None,
                                ret_checksum="checksum", exc_class=None):
        location = "file://foo/bar"
        checksum = "checksum"
        size = 10
        update_data = {'checksum': checksum}
        if ext_update_data is not None:
            update_data.update(ext_update_data)
        image_meta = {'id': unit_test_utils.UUID1,
                      'size': image_size}
        image_data = "blah"

        store = mock.MagicMock()
        notifier = mock.MagicMock()

        if exc_class is not None:
            store.add.side_effect = exc_class
        else:
            store.add.return_value = (location, size, ret_checksum, {})
        yield (location, checksum, image_meta, image_data, store, notifier,
               update_data)

    def test_upload_data_to_store(self):
        # 'user_storage_quota' is not set
        def store_add(image_id, data, size, **kwargs):
            # Check if 'data' is instance of 'CooperativeReader' when
            # 'user_storage_quota' is disabled.
            self.assertIsInstance(data, utils.CooperativeReader)
            return location, 10, "checksum", {}

        req = unit_test_utils.get_fake_request()
        with self._get_store_and_notifier(
                ext_update_data={'size': 10},
                exc_class=store_add) as (location, checksum, image_meta,
                                         image_data, store, notifier,
                                         update_data):
            ret = image_meta.update(update_data)
            with patch.object(registry, 'update_image_metadata',
                              return_value=ret) as mock_update_image_metadata:
                actual_meta, location_data = upload_utils.upload_data_to_store(
                    req, image_meta, image_data, store, notifier)

                self.assertEqual(location, location_data['url'])
                self.assertEqual(image_meta.update(update_data), actual_meta)
                mock_update_image_metadata.assert_called_once_with(
                    req.context, image_meta['id'], update_data,
                    from_state='saving')

    def test_upload_data_to_store_user_storage_quota_enabled(self):
        # Enable user_storage_quota
        self.config(user_storage_quota='100B')

        def store_add(image_id, data, size, **kwargs):
            # Check if 'data' is instance of 'LimitingReader' when
            # 'user_storage_quota' is enabled.
            self.assertIsInstance(data, utils.LimitingReader)
            return location, 10, "checksum", {}

        req = unit_test_utils.get_fake_request()
        with self._get_store_and_notifier(
                ext_update_data={'size': 10},
                exc_class=store_add) as (location, checksum, image_meta,
                                         image_data, store, notifier,
                                         update_data):
            ret = image_meta.update(update_data)
            # mock 'check_quota'
            mock_check_quota = patch('glance.api.common.check_quota',
                                     return_value=100)
            mock_check_quota.start()
            self.addCleanup(mock_check_quota.stop)
            with patch.object(registry, 'update_image_metadata',
                              return_value=ret) as mock_update_image_metadata:
                actual_meta, location_data = upload_utils.upload_data_to_store(
                    req, image_meta, image_data, store, notifier)

                self.assertEqual(location, location_data['url'])
                self.assertEqual(image_meta.update(update_data), actual_meta)
                mock_update_image_metadata.assert_called_once_with(
                    req.context, image_meta['id'], update_data,
                    from_state='saving')
                # 'check_quota' is called two times
                check_quota_call_count = (
                    mock_check_quota.target.check_quota.call_count)
                self.assertEqual(2, check_quota_call_count)

    def test_upload_data_to_store_mismatch_size(self):
        req = unit_test_utils.get_fake_request()

        with self._get_store_and_notifier(
            image_size=11) as (location, checksum, image_meta, image_data,
                               store, notifier, update_data):
            ret = image_meta.update(update_data)
            with patch.object(registry, 'update_image_metadata',
                              return_value=ret) as mock_update_image_metadata:
                self.assertRaises(webob.exc.HTTPBadRequest,
                                  upload_utils.upload_data_to_store,
                                  req, image_meta, image_data, store,
                                  notifier)
                mock_update_image_metadata.assert_called_with(
                    req.context, image_meta['id'], {'status': 'killed'},
                    from_state='saving')

    def test_upload_data_to_store_mismatch_checksum(self):
        req = unit_test_utils.get_fake_request()

        with self._get_store_and_notifier(
            ret_checksum='fake') as (location, checksum, image_meta,
                                     image_data, store, notifier, update_data):
            ret = image_meta.update(update_data)
            with patch.object(registry, "update_image_metadata",
                              return_value=ret) as mock_update_image_metadata:
                self.assertRaises(webob.exc.HTTPBadRequest,
                                  upload_utils.upload_data_to_store,
                                  req, image_meta, image_data, store,
                                  notifier)
                mock_update_image_metadata.assert_called_with(
                    req.context, image_meta['id'], {'status': 'killed'},
                    from_state='saving')

    def _test_upload_data_to_store_exception(self, exc_class, expected_class):
        req = unit_test_utils.get_fake_request()

        with self._get_store_and_notifier(
            exc_class=exc_class) as (location, checksum, image_meta,
                                     image_data, store, notifier, update_data):
            with patch.object(upload_utils, 'safe_kill') as mock_safe_kill:
                self.assertRaises(expected_class,
                                  upload_utils.upload_data_to_store,
                                  req, image_meta, image_data, store, notifier)
                mock_safe_kill.assert_called_once_with(
                    req, image_meta['id'], 'saving')

    def _test_upload_data_to_store_exception_with_notify(self,
                                                         exc_class,
                                                         expected_class,
                                                         image_killed=True):
        req = unit_test_utils.get_fake_request()

        with self._get_store_and_notifier(
            exc_class=exc_class) as (location, checksum, image_meta,
                                     image_data, store, notifier, update_data):
            with patch.object(upload_utils, 'safe_kill') as mock_safe_kill:
                self.assertRaises(expected_class,
                                  upload_utils.upload_data_to_store,
                                  req, image_meta, image_data, store,
                                  notifier)
                if image_killed:
                    mock_safe_kill.assert_called_with(req, image_meta['id'],
                                                      'saving')

    def test_upload_data_to_store_raises_store_disabled(self):
        """Test StoreDisabled exception is raised while uploading data"""
        self._test_upload_data_to_store_exception_with_notify(
            glance_store.StoreAddDisabled,
            webob.exc.HTTPGone,
            image_killed=True)

    def test_upload_data_to_store_duplicate(self):
        """See note in glance.api.v1.upload_utils on why we don't want image to
        be deleted in this case.
        """
        self._test_upload_data_to_store_exception_with_notify(
            exception.Duplicate,
            webob.exc.HTTPConflict,
            image_killed=False)

    def test_upload_data_to_store_forbidden(self):
        self._test_upload_data_to_store_exception_with_notify(
            exception.Forbidden,
            webob.exc.HTTPForbidden)

    def test_upload_data_to_store_storage_full(self):
        self._test_upload_data_to_store_exception_with_notify(
            glance_store.StorageFull,
            webob.exc.HTTPRequestEntityTooLarge)

    def test_upload_data_to_store_storage_write_denied(self):
        self._test_upload_data_to_store_exception_with_notify(
            glance_store.StorageWriteDenied,
            webob.exc.HTTPServiceUnavailable)

    def test_upload_data_to_store_size_limit_exceeded(self):
        self._test_upload_data_to_store_exception_with_notify(
            exception.ImageSizeLimitExceeded,
            webob.exc.HTTPRequestEntityTooLarge)

    def test_upload_data_to_store_http_error(self):
        self._test_upload_data_to_store_exception_with_notify(
            webob.exc.HTTPError,
            webob.exc.HTTPError)

    def test_upload_data_to_store_client_disconnect(self):
        self._test_upload_data_to_store_exception(
            ValueError,
            webob.exc.HTTPBadRequest)

    def test_upload_data_to_store_client_disconnect_ioerror(self):
        self._test_upload_data_to_store_exception(
            IOError,
            webob.exc.HTTPBadRequest)

    def test_upload_data_to_store_exception(self):
        self._test_upload_data_to_store_exception_with_notify(
            Exception,
            webob.exc.HTTPInternalServerError)

    def test_upload_data_to_store_not_found_after_upload(self):
        req = unit_test_utils.get_fake_request()

        with self._get_store_and_notifier(
            ext_update_data={'size': 10}) as (location, checksum, image_meta,
                                              image_data, store, notifier,
                                              update_data):
            exc = exception.ImageNotFound
            with patch.object(registry, 'update_image_metadata',
                              side_effect=exc) as mock_update_image_metadata:
                with patch.object(upload_utils,
                                  "initiate_deletion") as mock_initiate_del:
                    with patch.object(upload_utils,
                                      "safe_kill") as mock_safe_kill:
                        self.assertRaises(webob.exc.HTTPPreconditionFailed,
                                          upload_utils.upload_data_to_store,
                                          req, image_meta, image_data, store,
                                          notifier)
                        mock_update_image_metadata.assert_called_once_with(
                            req.context, image_meta['id'], update_data,
                            from_state='saving')
                        mock_initiate_del.assert_called_once_with(
                            req, {'url': location, 'status': 'active',
                                  'metadata': {}}, image_meta['id'])
                        mock_safe_kill.assert_called_once_with(
                            req, image_meta['id'], 'saving')

    @mock.patch.object(registry, 'update_image_metadata',
                       side_effect=exception.NotAuthenticated)
    @mock.patch.object(upload_utils, 'initiate_deletion')
    def test_activate_image_with_expired_token(
            self, mocked_delete, mocked_update):
        """Test token expiration during image upload.

        If users token expired before image was uploaded then if auth error
        was caught from registry during changing image status from 'saving'
        to 'active' then it's required to delete all image data.
        """
        context = mock.Mock()
        req = mock.Mock()
        req.context = context
        with self._get_store_and_notifier() as (location, checksum, image_meta,
                                                image_data, store, notifier,
                                                update_data):
            self.assertRaises(webob.exc.HTTPUnauthorized,
                              upload_utils.upload_data_to_store,
                              req, image_meta, image_data, store, notifier)
            self.assertEqual(2, mocked_update.call_count)
            mocked_delete.assert_called_once_with(
                req,
                {'url': 'file://foo/bar', 'status': 'active', 'metadata': {}},
                'c80a1a6c-bd1f-41c5-90ee-81afedb1d58d')



# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import numbers
import re

import semantic_version
import six

import glance.common.exception as exc
from glance.common.glare import declarative
from glance.i18n import _


class Text(declarative.PropertyDefinition):
    """A text metadata property of arbitrary length

    Maps to TEXT columns in database, does not support sorting or filtering
    """
    ALLOWED_TYPES = (six.string_types,)
    DB_TYPE = 'text'


# noinspection PyAttributeOutsideInit
class String(Text):
    """A string metadata property of limited length

    Maps to VARCHAR columns in database, supports filtering and sorting.
    May have constrains on length and regexp patterns.

    The maximum length is limited to 255 characters
    """

    DB_TYPE = 'string'

    def __init__(self, max_length=255, min_length=0, pattern=None, **kwargs):
        """Defines a String metadata property.

        :param max_length: maximum value length
        :param min_length: minimum value length
        :param pattern: regexp pattern to match
        """
        super(String, self).__init__(**kwargs)

        self.max_length(max_length)
        self.min_length(min_length)
        if pattern:
            self.pattern(pattern)
        # if default and/or allowed_values are specified (in base classes)
        # then we need to validate them against the newly added validators
        self._check_definition()

    def max_length(self, value):
        """Sets the maximum value length"""
        self._max_length = value
        if value is not None:
            if value > 255:
                raise exc.InvalidArtifactTypePropertyDefinition(
                    _('Max string length may not exceed 255 characters'))
            self._add_validator('max_length',
                                lambda v: len(v) <= self._max_length,
                                _('Length  is greater than maximum'))
        else:
            self._remove_validator('max_length')
        self._check_definition()

    def min_length(self, value):
        """Sets the minimum value length"""
        self._min_length = value
        if value is not None:
            if value < 0:
                raise exc.InvalidArtifactTypePropertyDefinition(
                    _('Min string length may not be negative'))

            self._add_validator('min_length',
                                lambda v: len(v) >= self._min_length,
                                _('Length is less than minimum'))
        else:
            self._remove_validator('min_length')
        self._check_definition()

    def pattern(self, value):
        """Sets the regexp pattern to match"""
        self._pattern = value
        if value is not None:
            self._add_validator('pattern',
                                lambda v: re.match(self._pattern,
                                                   v) is not None,
                                _('Does not match pattern'))
        else:
            self._remove_validator('pattern')
        self._check_definition()


class SemVerString(String):
    """A String metadata property matching semver pattern"""

    def __init__(self, **kwargs):
        def validate(value):
            try:
                semantic_version.Version(value, partial=True)
            except ValueError:
                return False
            return True

        super(SemVerString,
              self).__init__(validators=[(validate,
                                         "Invalid semver string")],
                             **kwargs)


# noinspection PyAttributeOutsideInit
class Integer(declarative.PropertyDefinition):
    """An Integer metadata property

    Maps to INT columns in Database, supports filtering and sorting.
    May have constraints on value
    """

    ALLOWED_TYPES = (six.integer_types,)
    DB_TYPE = 'int'

    def __init__(self, min_value=None, max_value=None, **kwargs):
        """Defines an Integer metadata property

        :param min_value: minimum allowed value
        :param max_value: maximum allowed value
        """
        super(Integer, self).__init__(**kwargs)
        if min_value is not None:
            self.min_value(min_value)

        if max_value is not None:
            self.max_value(max_value)

        # if default and/or allowed_values are specified (in base classes)
        # then we need to validate them against the newly added validators
        self._check_definition()

    def min_value(self, value):
        """Sets the minimum allowed value"""
        self._min_value = value
        if value is not None:
            self._add_validator('min_value',
                                lambda v: v >= self._min_value,
                                _('Value is less than minimum'))
        else:
            self._remove_validator('min_value')
        self._check_definition()

    def max_value(self, value):
        """Sets the maximum allowed value"""
        self._max_value = value
        if value is not None:
            self._add_validator('max_value',
                                lambda v: v <= self._max_value,
                                _('Value is greater than maximum'))
        else:
            self._remove_validator('max_value')
        self._check_definition()


# noinspection PyAttributeOutsideInit
class DateTime(declarative.PropertyDefinition):
    """A DateTime metadata property

    Maps to a DATETIME columns in database.
    Is not supported as Type Specific property, may be used only as Generic one

    May have constraints on value
    """
    ALLOWED_TYPES = (datetime.datetime,)
    DB_TYPE = 'datetime'

    def __init__(self, min_value=None, max_value=None, **kwargs):
        """Defines a DateTime metadata property

        :param min_value: minimum allowed value
        :param max_value: maximum allowed value
        """
        super(DateTime, self).__init__(**kwargs)
        if min_value is not None:
            self.min_value(min_value)

        if max_value is not None:
            self.max_value(max_value)

        # if default and/or allowed_values are specified (in base classes)
        # then we need to validate them against the newly added validators
        self._check_definition()

    def min_value(self, value):
        """Sets the minimum allowed value"""
        self._min_value = value
        if value is not None:
            self._add_validator('min_value',
                                lambda v: v >= self._min_value,
                                _('Value is less than minimum'))
        else:
            self._remove_validator('min_value')
        self._check_definition()

    def max_value(self, value):
        """Sets the maximum allowed value"""
        self._max_value = value
        if value is not None:
            self._add_validator('max_value',
                                lambda v: v <= self._max_value,
                                _('Value is greater than maximum'))
        else:
            self._remove_validator('max_value')
        self._check_definition()


# noinspection PyAttributeOutsideInit
class Numeric(declarative.PropertyDefinition):
    """A Numeric metadata property

    Maps to floating point number columns in Database, supports filtering and
    sorting. May have constraints on value
    """
    ALLOWED_TYPES = numbers.Number
    DB_TYPE = 'numeric'

    def __init__(self, min_value=None, max_value=None, **kwargs):
        """Defines a Numeric metadata property

        :param min_value: minimum allowed value
        :param max_value: maximum allowed value
        """
        super(Numeric, self).__init__(**kwargs)
        if min_value is not None:
            self.min_value(min_value)

        if max_value is not None:
            self.max_value(max_value)

        # if default and/or allowed_values are specified (in base classes)
        # then we need to validate them against the newly added validators
        self._check_definition()

    def min_value(self, value):
        """Sets the minimum allowed value"""
        self._min_value = value
        if value is not None:
            self._add_validator('min_value',
                                lambda v: v >= self._min_value,
                                _('Value is less than minimum'))
        else:
            self._remove_validator('min_value')
        self._check_definition()

    def max_value(self, value):
        """Sets the maximum allowed value"""
        self._max_value = value
        if value is not None:
            self._add_validator('max_value',
                                lambda v: v <= self._max_value,
                                _('Value is greater than maximum'))
        else:
            self._remove_validator('max_value')
        self._check_definition()


class Boolean(declarative.PropertyDefinition):
    """A Boolean metadata property

    Maps to Boolean columns in database. Supports filtering and sorting.
    """
    ALLOWED_TYPES = (bool,)
    DB_TYPE = 'bool'


class Array(declarative.ListAttributeDefinition,
            declarative.PropertyDefinition, list):
    """An array metadata property

    May contain elements of any other PropertyDefinition types except Dict and
    Array. Each elements maps to appropriate type of columns in database.
    Preserves order. Allows filtering based on "Array contains Value" semantics

    May specify constrains on types of elements, their amount and uniqueness.
    """
    ALLOWED_ITEM_TYPES = (declarative.PropertyDefinition,)

    def __init__(self, item_type=String(), min_size=0, max_size=None,
                 unique=False, extra_items=True, **kwargs):
        """Defines an Array metadata property

        :param item_type: defines the types of elements in Array. If set to an
        instance of PropertyDefinition then all the elements have to be of that
        type. If set to list of such instances, then the elements on the
        corresponding positions have to be of the appropriate type.
        :param min_size: minimum size of the Array
        :param max_size: maximum size of the Array
        :param unique: if set to true, all the elements in the Array have to be
        unique
        """
        if isinstance(item_type, Array):
            msg = _("Array property can't have item_type=Array")
            raise exc.InvalidArtifactTypePropertyDefinition(msg)
        declarative.ListAttributeDefinition.__init__(self,
                                                     item_type=item_type,
                                                     min_size=min_size,
                                                     max_size=max_size,
                                                     unique=unique)
        declarative.PropertyDefinition.__init__(self, **kwargs)


class Dict(declarative.DictAttributeDefinition,
           declarative.PropertyDefinition, dict):
    """A dictionary metadata property

    May contain elements of any other PropertyDefinition types except Dict.
    Each elements maps to appropriate type of columns in database. Allows
    filtering and sorting by values of each key except the ones mapping the
    Text fields.

    May specify constrains on types of elements and their amount.
    """
    ALLOWED_PROPERTY_TYPES = (declarative.PropertyDefinition,)

    def __init__(self, properties=String(), min_properties=0,
                 max_properties=None, **kwargs):
        """Defines a dictionary metadata property

        :param properties: defines the types of dictionary values. If set to an
        instance of PropertyDefinition then all the value have to be of that
        type. If set to a dictionary with string keys and values of
        PropertyDefinition type, then the elements mapped by the corresponding
        have have to be of the appropriate type.
        :param min_properties: minimum allowed amount of properties in the dict
        :param max_properties: maximum allowed amount of properties in the dict
        """
        declarative.DictAttributeDefinition.__init__(
            self,
            properties=properties,
            min_properties=min_properties,
            max_properties=max_properties)
        declarative.PropertyDefinition.__init__(self, **kwargs)


class ArtifactType(declarative.get_declarative_base()):  # noqa
    """A base class for all the Artifact Type definitions

    Defines the Generic metadata properties as attributes.
    """
    id = String(required=True, readonly=True)
    type_name = String(required=True, readonly=True)
    type_version = SemVerString(required=True, readonly=True)
    name = String(required=True, mutable=False)
    version = SemVerString(required=True, mutable=False)
    description = Text()
    tags = Array(unique=True, default=[])
    visibility = String(required=True,
                        allowed_values=["private", "public", "shared",
                                        "community"],
                        default="private")
    state = String(required=True, readonly=True, allowed_values=["creating",
                                                                 "active",
                                                                 "deactivated",
                                                                 "deleted"])
    owner = String(required=True, readonly=True)
    created_at = DateTime(required=True, readonly=True)
    updated_at = DateTime(required=True, readonly=True)
    published_at = DateTime(readonly=True)
    deleted_at = DateTime(readonly=True)

    def __init__(self, **kwargs):
        if "type_name" in kwargs:
            raise exc.InvalidArtifactPropertyValue(
                _("Unable to specify artifact type explicitly"))
        if "type_version" in kwargs:
            raise exc.InvalidArtifactPropertyValue(
                _("Unable to specify artifact type version explicitly"))
        super(ArtifactType,
              self).__init__(type_name=self.metadata.type_name,
                             type_version=self.metadata.type_version, **kwargs)

    def __eq__(self, other):
        if not isinstance(other, ArtifactType):
            return False
        return self.id == other.id

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.id)

    def __is_mutable__(self):
        return self.state == "creating"


class ArtifactReference(declarative.RelationDefinition):
    """An artifact reference definition

    Allows to define constraints by the name and version of target artifact
    """
    ALLOWED_TYPES = ArtifactType

    def __init__(self, type_name=None, type_version=None, **kwargs):
        """Defines an artifact reference

        :param type_name: type name of the target artifact
        :param type_version: type version of the target artifact
        """
        super(ArtifactReference, self).__init__(**kwargs)
        if type_name is not None:
            if isinstance(type_name, list):
                type_names = list(type_name)
                if type_version is not None:
                    raise exc.InvalidArtifactTypePropertyDefinition(
                        _('Unable to specify version '
                          'if multiple types are possible'))
            else:
                type_names = [type_name]

            def validate_reference(artifact):
                if artifact.type_name not in type_names:
                    return False
                if (type_version is not None and
                   artifact.type_version != type_version):
                    return False
                return True

            self._add_validator('referenced_type',
                                validate_reference,
                                _("Invalid referenced type"))
        elif type_version is not None:
            raise exc.InvalidArtifactTypePropertyDefinition(
                _('Unable to specify version '
                  'if type is not specified'))
        self._check_definition()


class ArtifactReferenceList(declarative.ListAttributeDefinition,
                            declarative.RelationDefinition, list):
    """A list of Artifact References

    Allows to define a collection of references to other artifacts, each
    optionally constrained by type name and type version
    """
    ALLOWED_ITEM_TYPES = (ArtifactReference,)

    def __init__(self, references=ArtifactReference(), min_size=0,
                 max_size=None, **kwargs):
        if isinstance(references, list):
            raise exc.InvalidArtifactTypePropertyDefinition(
                _("Invalid reference list specification"))
        declarative.RelationDefinition.__init__(self, **kwargs)
        declarative.ListAttributeDefinition.__init__(self,
                                                     item_type=references,
                                                     min_size=min_size,
                                                     max_size=max_size,
                                                     unique=True,
                                                     default=[]
                                                     if min_size == 0 else
                                                     None)


class Blob(object):
    """A Binary object being part of the Artifact"""
    def __init__(self, size=0, locations=None, checksum=None, item_key=None):
        """Initializes a new Binary Object for an Artifact

        :param size: the size of Binary Data
        :param locations: a list of data locations in backing stores
        :param checksum: a checksum for the data
        """
        if locations is None:
            locations = []
        self.size = size
        self.checksum = checksum
        self.locations = locations
        self.item_key = item_key

    def to_dict(self):
        return {
            "size": self.size,
            "checksum": self.checksum,
        }


class BinaryObject(declarative.BlobDefinition, Blob):
    """A definition of BinaryObject binding

    Adds a BinaryObject to an Artifact Type, optionally constrained by file
    size and amount of locations
    """
    ALLOWED_TYPES = (Blob,)

    def __init__(self,
                 max_file_size=None,
                 min_file_size=None,
                 min_locations=None,
                 max_locations=None,
                 **kwargs):
        """Defines a binary object as part of Artifact Type
        :param max_file_size: maximum size of the associate Blob
        :param min_file_size: minimum size of the associated Blob
        :param min_locations: minimum number of locations in the associated
        Blob
        :param max_locations: maximum number of locations in the associated
        Blob
        """
        mutable = kwargs.pop('mutable', False)
        if mutable:
            raise exc.InvalidArtifactTypePropertyDefinition(
                _("BinaryObject property cannot be declared mutable"))
        super(BinaryObject, self).__init__(default=None, readonly=False,
                                           mutable=mutable, **kwargs)
        self._max_file_size = max_file_size
        self._min_file_size = min_file_size
        self._min_locations = min_locations
        self._max_locations = max_locations

        self._add_validator('size_not_empty',
                            lambda v: v.size is not None,
                            _('Blob size is not set'))
        if max_file_size:
            self._add_validator('max_size',
                                lambda v: v.size <= self._max_file_size,
                                _("File too large"))
        if min_file_size:
            self._add_validator('min_size',
                                lambda v: v.size >= self._min_file_size,
                                _("File too small"))
        if min_locations:
            self._add_validator('min_locations',
                                lambda v: len(
                                    v.locations) >= self._min_locations,
                                _("Too few locations"))
        if max_locations:
            self._add_validator(
                'max_locations',
                lambda v: len(v.locations) <= self._max_locations,
                _("Too many locations"))


class BinaryObjectList(declarative.ListAttributeDefinition,
                       declarative.BlobDefinition, list):
    """A definition of binding to the list of BinaryObject

    Adds a list of BinaryObject's to an artifact type, optionally constrained
    by the number of objects in the list and their uniqueness

    """
    ALLOWED_ITEM_TYPES = (BinaryObject,)

    def __init__(self, objects=BinaryObject(), min_count=0, max_count=None,
                 **kwargs):
        declarative.BlobDefinition.__init__(self, **kwargs)
        declarative.ListAttributeDefinition.__init__(self,
                                                     item_type=objects,
                                                     min_size=min_count,
                                                     max_size=max_count,
                                                     unique=True)
        self.default = [] if min_count == 0 else None



# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from glance.common import exception
from glance.common import semver_db
from glance.tests import utils as test_utils


class SemVerTestCase(test_utils.BaseTestCase):
    def test_long_conversion(self):
        initial = '1.2.3-beta+07.17.2014'
        v = semver_db.parse(initial)
        l, prerelease, build = v.__composite_values__()
        v2 = semver_db.DBVersion(l, prerelease, build)
        self.assertEqual(initial, str(v2))

    def test_major_comparison_as_long(self):
        v1 = semver_db.parse("1.1.100")
        v2 = semver_db.parse("2.0.0")
        self.assertTrue(v2.__composite_values__()[0] >
                        v1.__composite_values__()[0])

    def test_minor_comparison_as_long(self):
        v1 = semver_db.parse("1.1.100")
        v2 = semver_db.parse("2.0.0")
        self.assertTrue(v2.__composite_values__()[0] >
                        v1.__composite_values__()[0])

    def test_patch_comparison_as_long(self):
        v1 = semver_db.parse("1.1.1")
        v2 = semver_db.parse("1.1.100")
        self.assertTrue(v2.__composite_values__()[0] >
                        v1.__composite_values__()[0])

    def test_label_comparison_as_long(self):
        v1 = semver_db.parse("1.1.1-alpha")
        v2 = semver_db.parse("1.1.1")
        self.assertTrue(v2.__composite_values__()[0] >
                        v1.__composite_values__()[0])

    def test_label_comparison_as_string(self):
        versions = [
            semver_db.parse("1.1.1-0.10.a.23.y.255").__composite_values__()[1],
            semver_db.parse("1.1.1-0.10.z.23.x.255").__composite_values__()[1],
            semver_db.parse("1.1.1-0.10.z.23.y.255").__composite_values__()[1],
            semver_db.parse("1.1.1-0.10.z.23.y.256").__composite_values__()[1],
            semver_db.parse("1.1.1-0.10.z.24.y.255").__composite_values__()[1],
            semver_db.parse("1.1.1-0.11.z.24.y.255").__composite_values__()[1],
            semver_db.parse("1.1.1-1.11.z.24.y.255").__composite_values__()[1],
            semver_db.parse("1.1.1-alp.1.2.3.4.5.6").__composite_values__()[1]]
        for i in range(len(versions) - 1):
            self.assertLess(versions[i], versions[i + 1])

    def test_too_large_version(self):
        version1 = '1.1.65536'
        version2 = '1.65536.1'
        version3 = '65536.1.1'
        self.assertRaises(exception.InvalidVersion, semver_db.parse, version1)
        self.assertRaises(exception.InvalidVersion, semver_db.parse, version2)
        self.assertRaises(exception.InvalidVersion, semver_db.parse, version3)

    def test_too_long_numeric_segments(self):
        version = semver_db.parse('1.0.0-alpha.1234567')
        self.assertRaises(exception.InvalidVersion,
                          version.__composite_values__)



# Copyright 2011-2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import sys
import time

import httplib2
from oslo_serialization import jsonutils
from oslo_utils import units
# NOTE(jokke): simplified transition to py3, behaves like py2 xrange
from six.moves import range

from glance.tests import functional
from glance.tests.utils import execute


TEST_IMAGE_DATA = '*' * 5 * units.Ki
TEST_IMAGE_META = {
    'name': 'test_image',
    'is_public': False,
    'disk_format': 'raw',
    'container_format': 'ovf',
}


class TestScrubber(functional.FunctionalTest):

    """Test that delayed_delete works and the scrubber deletes"""

    def _send_http_request(self, path, method, body=None):
        headers = {
            'x-image-meta-name': 'test_image',
            'x-image-meta-is_public': 'true',
            'x-image-meta-disk_format': 'raw',
            'x-image-meta-container_format': 'ovf',
            'content-type': 'application/octet-stream'
        }
        return httplib2.Http().request(path, method, body, headers)

    def test_delayed_delete(self):
        """
        test that images don't get deleted immediately and that the scrubber
        scrubs them
        """
        self.cleanup()
        self.start_servers(delayed_delete=True, daemon=True,
                           metadata_encryption_key='')
        path = "http://%s:%d/v1/images" % ("127.0.0.1", self.api_port)
        response, content = self._send_http_request(path, 'POST', body='XXX')
        self.assertEqual(201, response.status)
        image = jsonutils.loads(content)['image']
        self.assertEqual('active', image['status'])

        path = "http://%s:%d/v1/images/%s" % ("127.0.0.1", self.api_port,
                                              image['id'])
        response, content = self._send_http_request(path, 'DELETE')
        self.assertEqual(200, response.status)
        response, content = self._send_http_request(path, 'HEAD')
        self.assertEqual(200, response.status)
        self.assertEqual('pending_delete', response['x-image-meta-status'])

        self.wait_for_scrub(path)

        self.stop_servers()

    def test_delayed_delete_with_trustedauth_registry(self):
        """
        test that images don't get deleted immediately and that the scrubber
        scrubs them when registry is operating in trustedauth mode
        """
        self.cleanup()
        self.api_server.deployment_flavor = 'noauth'
        self.registry_server.deployment_flavor = 'trusted-auth'
        self.start_servers(delayed_delete=True, daemon=True,
                           metadata_encryption_key='',
                           send_identity_headers=True)
        base_headers = {
            'X-Identity-Status': 'Confirmed',
            'X-Auth-Token': '932c5c84-02ac-4fe5-a9ba-620af0e2bb96',
            'X-User-Id': 'f9a41d13-0c13-47e9-bee2-ce4e8bfe958e',
            'X-Tenant-Id': 'deae8923-075d-4287-924b-840fb2644874',
            'X-Roles': 'admin',
        }
        headers = {
            'x-image-meta-name': 'test_image',
            'x-image-meta-is_public': 'true',
            'x-image-meta-disk_format': 'raw',
            'x-image-meta-container_format': 'ovf',
            'content-type': 'application/octet-stream',
        }
        headers.update(base_headers)
        path = "http://%s:%d/v1/images" % ("127.0.0.1", self.api_port)
        http = httplib2.Http()
        response, content = http.request(path, 'POST', body='XXX',
                                         headers=headers)
        self.assertEqual(201, response.status)
        image = jsonutils.loads(content)['image']
        self.assertEqual('active', image['status'])
        image_id = image['id']

        path = "http://%s:%d/v1/images/%s" % ("127.0.0.1", self.api_port,
                                              image_id)
        http = httplib2.Http()
        response, content = http.request(path, 'DELETE', headers=base_headers)
        self.assertEqual(200, response.status)

        response, content = http.request(path, 'HEAD', headers=base_headers)
        self.assertEqual(200, response.status)
        self.assertEqual('pending_delete', response['x-image-meta-status'])

        self.wait_for_scrub(path, headers=base_headers)

        self.stop_servers()

    def test_scrubber_app(self):
        """
        test that the glance-scrubber script runs successfully when not in
        daemon mode
        """
        self.cleanup()
        self.start_servers(delayed_delete=True, daemon=False,
                           metadata_encryption_key='')

        path = "http://%s:%d/v1/images" % ("127.0.0.1", self.api_port)
        response, content = self._send_http_request(path, 'POST', body='XXX')
        self.assertEqual(201, response.status)
        image = jsonutils.loads(content)['image']
        self.assertEqual('active', image['status'])

        path = "http://%s:%d/v1/images/%s" % ("127.0.0.1", self.api_port,
                                              image['id'])
        response, content = self._send_http_request(path, 'DELETE')
        self.assertEqual(200, response.status)

        response, content = self._send_http_request(path, 'HEAD')
        self.assertEqual(200, response.status)
        self.assertEqual('pending_delete', response['x-image-meta-status'])

        # wait for the scrub time on the image to pass
        time.sleep(self.api_server.scrub_time)

        # scrub images and make sure they get deleted
        exe_cmd = "%s -m glance.cmd.scrubber" % sys.executable
        cmd = ("%s --config-file %s" %
               (exe_cmd, self.scrubber_daemon.conf_file_name))
        exitcode, out, err = execute(cmd, raise_error=False)
        self.assertEqual(0, exitcode)

        self.wait_for_scrub(path)

        self.stop_servers()

    def test_scrubber_app_with_trustedauth_registry(self):
        """
        test that the glance-scrubber script runs successfully when not in
        daemon mode and with a registry that operates in trustedauth mode
        """
        self.cleanup()
        self.api_server.deployment_flavor = 'noauth'
        self.registry_server.deployment_flavor = 'trusted-auth'
        self.start_servers(delayed_delete=True, daemon=False,
                           metadata_encryption_key='',
                           send_identity_headers=True)
        base_headers = {
            'X-Identity-Status': 'Confirmed',
            'X-Auth-Token': '932c5c84-02ac-4fe5-a9ba-620af0e2bb96',
            'X-User-Id': 'f9a41d13-0c13-47e9-bee2-ce4e8bfe958e',
            'X-Tenant-Id': 'deae8923-075d-4287-924b-840fb2644874',
            'X-Roles': 'admin',
        }
        headers = {
            'x-image-meta-name': 'test_image',
            'x-image-meta-is_public': 'true',
            'x-image-meta-disk_format': 'raw',
            'x-image-meta-container_format': 'ovf',
            'content-type': 'application/octet-stream',
        }
        headers.update(base_headers)
        path = "http://%s:%d/v1/images" % ("127.0.0.1", self.api_port)
        http = httplib2.Http()
        response, content = http.request(path, 'POST', body='XXX',
                                         headers=headers)
        self.assertEqual(201, response.status)
        image = jsonutils.loads(content)['image']
        self.assertEqual('active', image['status'])
        image_id = image['id']

        path = "http://%s:%d/v1/images/%s" % ("127.0.0.1", self.api_port,
                                              image_id)
        http = httplib2.Http()
        response, content = http.request(path, 'DELETE', headers=base_headers)
        self.assertEqual(200, response.status)

        response, content = http.request(path, 'HEAD', headers=base_headers)
        self.assertEqual(200, response.status)
        self.assertEqual('pending_delete', response['x-image-meta-status'])

        # wait for the scrub time on the image to pass
        time.sleep(self.api_server.scrub_time)

        # scrub images and make sure they get deleted
        exe_cmd = "%s -m glance.cmd.scrubber" % sys.executable
        cmd = ("%s --config-file %s" %
               (exe_cmd, self.scrubber_daemon.conf_file_name))
        exitcode, out, err = execute(cmd, raise_error=False)
        self.assertEqual(0, exitcode)

        self.wait_for_scrub(path, headers=base_headers)

        self.stop_servers()

    def test_scrubber_delete_handles_exception(self):
        """
        Test that the scrubber handles the case where an
        exception occurs when _delete() is called. The scrubber
        should not write out queue files in this case.
        """

        # Start servers.
        self.cleanup()
        self.start_servers(delayed_delete=True, daemon=False,
                           default_store='file')

        # Check that we are using a file backend.
        self.assertEqual(self.api_server.default_store, 'file')

        # add an image
        path = "http://%s:%d/v1/images" % ("127.0.0.1", self.api_port)
        response, content = self._send_http_request(path, 'POST', body='XXX')
        self.assertEqual(201, response.status)
        image = jsonutils.loads(content)['image']
        self.assertEqual('active', image['status'])

        # delete the image
        path = "http://%s:%d/v1/images/%s" % ("127.0.0.1", self.api_port,
                                              image['id'])
        response, content = self._send_http_request(path, 'DELETE')
        self.assertEqual(200, response.status)

        # ensure the image is marked pending delete
        response, content = self._send_http_request(path, 'HEAD')
        self.assertEqual(200, response.status)
        self.assertEqual('pending_delete', response['x-image-meta-status'])

        # Remove the file from the backend.
        file_path = os.path.join(self.api_server.image_dir, image['id'])
        os.remove(file_path)

        # Wait for the scrub time on the image to pass
        time.sleep(self.api_server.scrub_time)

        # run the scrubber app, and ensure it doesn't fall over
        exe_cmd = "%s -m glance.cmd.scrubber" % sys.executable
        cmd = ("%s --config-file %s" %
               (exe_cmd, self.scrubber_daemon.conf_file_name))
        exitcode, out, err = execute(cmd, raise_error=False)
        self.assertEqual(0, exitcode)

        self.wait_for_scrub(path)

        self.stop_servers()

    def wait_for_scrub(self, path, headers=None):
        """
        NOTE(jkoelker) The build servers sometimes take longer than 15 seconds
        to scrub. Give it up to 5 min, checking checking every 15 seconds.
        When/if it flips to deleted, bail immediately.
        """
        http = httplib2.Http()
        wait_for = 300    # seconds
        check_every = 15  # seconds
        for _ in range(wait_for / check_every):
            time.sleep(check_every)

            response, content = http.request(path, 'HEAD', headers=headers)
            if (response['x-image-meta-status'] == 'deleted' and
                    response['x-image-meta-deleted'] == 'True'):
                break
            else:
                continue
        else:
            self.fail('image was never scrubbed')



# Copyright 2010 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import calendar
import time

import eventlet
from glance_store import exceptions as store_exceptions
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import encodeutils
import six

from glance.common import crypt
from glance.common import exception
from glance import context
import glance.db as db_api
from glance.i18n import _, _LE, _LI, _LW
import glance.registry.client.v1.api as registry

LOG = logging.getLogger(__name__)

scrubber_opts = [
    cfg.IntOpt('scrub_time', default=0, min=0,
               help=_("""
The amount of time, in seconds, to delay image scrubbing.

When delayed delete is turned on, an image is put into ``pending_delete``
state upon deletion until the scrubber deletes its image data. Typically, soon
after the image is put into ``pending_delete`` state, it is available for
scrubbing. However, scrubbing can be delayed until a later point using this
configuration option. This option denotes the time period an image spends in
``pending_delete`` state before it is available for scrubbing.

It is important to realize that this has storage implications. The larger the
``scrub_time``, the longer the time to reclaim backend storage from deleted
images.

Possible values:
    * Any non-negative integer

Related options:
    * ``delayed_delete``

""")),
    cfg.IntOpt('scrub_pool_size', default=1, min=1,
               help=_("""
The size of thread pool to be used for scrubbing images.

When there are a large number of images to scrub, it is beneficial to scrub
images in parallel so that the scrub queue stays in control and the backend
storage is reclaimed in a timely fashion. This configuration option denotes
the maximum number of images to be scrubbed in parallel. The default value is
one, which signifies serial scrubbing. Any value above one indicates parallel
scrubbing.

Possible values:
    * Any non-zero positive integer

Related options:
    * ``delayed_delete``

""")),
    cfg.BoolOpt('delayed_delete', default=False,
                help=_("""
Turn on/off delayed delete.

Typically when an image is deleted, the ``glance-api`` service puts the image
into ``deleted`` state and deletes its data at the same time. Delayed delete
is a feature in Glance that delays the actual deletion of image data until a
later point in time (as determined by the configuration option ``scrub_time``).
When delayed delete is turned on, the ``glance-api`` service puts the image
into ``pending_delete`` state upon deletion and leaves the image data in the
storage backend for the image scrubber to delete at a later time. The image
scrubber will move the image into ``deleted`` state upon successful deletion
of image data.

NOTE: When delayed delete is turned on, image scrubber MUST be running as a
periodic task to prevent the backend storage from filling up with undesired
usage.

Possible values:
    * True
    * False

Related options:
    * ``scrub_time``
    * ``wakeup_time``
    * ``scrub_pool_size``

""")),

    # Note: Though the conf option admin_role is used by other Glance
    # service and their usage differs requiring us to have a differing
    # help text here, oslo.config generator treats them as the same
    # config option and would throw a DuplicateError exception in case
    # of differing help texts. Hence we have the same help text for
    # admin_role here and in context.py.

    cfg.StrOpt('admin_role', default='admin',
               help=_("""
Role used to identify an authenticated user as administrator.

Provide a string value representing a Keystone role to identify an
administrative user. Users with this role will be granted
administrative privileges. The default value for this option is
'admin'.

Possible values:
    * A string value which is a valid Keystone role

Related options:
    * None

""")),
    cfg.BoolOpt('send_identity_headers',
                default=False,
                help=_("""
Send headers received from identity when making requests to
registry.

Typically, Glance registry can be deployed in multiple flavors,
which may or may not include authentication. For example,
``trusted-auth`` is a flavor that does not require the registry
service to authenticate the requests it receives. However, the
registry service may still need a user context to be populated to
serve the requests. This can be achieved by the caller
(the Glance API usually) passing through the headers it received
from authenticating with identity for the same request. The typical
headers sent are ``X-User-Id``, ``X-Tenant-Id``, ``X-Roles``,
``X-Identity-Status`` and ``X-Service-Catalog``.

Provide a boolean value to determine whether to send the identity
headers to provide tenant and user information along with the
requests to registry service. By default, this option is set to
``False``, which means that user and tenant information is not
available readily. It must be obtained by authenticating. Hence, if
this is set to ``False``, ``flavor`` must be set to value that
either includes authentication or authenticated user context.

Possible values:
    * True
    * False

Related options:
    * flavor

""")),
]

scrubber_cmd_opts = [
    cfg.IntOpt('wakeup_time', default=300, min=0,
               help=_("""
Time interval, in seconds, between scrubber runs in daemon mode.

Scrubber can be run either as a cron job or daemon. When run as a daemon, this
configuration time specifies the time period between two runs. When the
scrubber wakes up, it fetches and scrubs all ``pending_delete`` images that
are available for scrubbing after taking ``scrub_time`` into consideration.

If the wakeup time is set to a large number, there may be a large number of
images to be scrubbed for each run. Also, this impacts how quickly the backend
storage is reclaimed.

Possible values:
    * Any non-negative integer

Related options:
    * ``daemon``
    * ``delayed_delete``

"""))
]

scrubber_cmd_cli_opts = [
    cfg.BoolOpt('daemon',
                short='D',
                default=False,
                help=_("""
Run scrubber as a daemon.

This boolean configuration option indicates whether scrubber should
run as a long-running process that wakes up at regular intervals to
scrub images. The wake up interval can be specified using the
configuration option ``wakeup_time``.

If this configuration option is set to ``False``, which is the
default value, scrubber runs once to scrub images and exits. In this
case, if the operator wishes to implement continuous scrubbing of
images, scrubber needs to be scheduled as a cron job.

Possible values:
    * True
    * False

Related options:
    * ``wakeup_time``

"""))
]

CONF = cfg.CONF
CONF.register_opts(scrubber_opts)
CONF.import_opt('metadata_encryption_key', 'glance.common.config')


class ScrubDBQueue(object):
    """Database-based image scrub queue class."""
    def __init__(self):
        self.scrub_time = CONF.scrub_time
        self.metadata_encryption_key = CONF.metadata_encryption_key
        registry.configure_registry_client()
        registry.configure_registry_admin_creds()
        admin_user = CONF.admin_user
        admin_tenant = CONF.admin_tenant_name

        if CONF.send_identity_headers:
            # When registry is operating in trusted-auth mode
            roles = [CONF.admin_role]
            self.admin_context = context.RequestContext(user=admin_user,
                                                        tenant=admin_tenant,
                                                        auth_token=None,
                                                        roles=roles)
            self.registry = registry.get_registry_client(self.admin_context)
        else:
            ctxt = context.RequestContext()
            self.registry = registry.get_registry_client(ctxt)
            admin_token = self.registry.auth_token
            self.admin_context = context.RequestContext(user=admin_user,
                                                        tenant=admin_tenant,
                                                        auth_token=admin_token)

    def add_location(self, image_id, location):
        """Adding image location to scrub queue.

        :param image_id: The opaque image identifier
        :param location: The opaque image location

        :returns: A boolean value to indicate success or not
        """
        loc_id = location.get('id')
        if loc_id:
            db_api.get_api().image_location_delete(self.admin_context,
                                                   image_id, loc_id,
                                                   'pending_delete')
            return True
        else:
            return False

    def _get_images_page(self, marker):
        filters = {'deleted': True,
                   'is_public': 'none',
                   'status': 'pending_delete'}

        if marker:
            return self.registry.get_images_detailed(filters=filters,
                                                     marker=marker)
        else:
            return self.registry.get_images_detailed(filters=filters)

    def _get_all_images(self):
        """Generator to fetch all appropriate images, paging as needed."""

        marker = None
        while True:
            images = self._get_images_page(marker)
            if len(images) == 0:
                break
            marker = images[-1]['id']

            for image in images:
                yield image

    def get_all_locations(self):
        """Returns a list of image id and location tuple from scrub queue.

        :returns: a list of image id, location id and uri tuple from
            scrub queue

        """
        ret = []

        for image in self._get_all_images():
            deleted_at = image.get('deleted_at')
            if not deleted_at:
                continue

            # NOTE: Strip off microseconds which may occur after the last '.,'
            # Example: 2012-07-07T19:14:34.974216
            date_str = deleted_at.rsplit('.', 1)[0].rsplit(',', 1)[0]
            delete_time = calendar.timegm(time.strptime(date_str,
                                                        "%Y-%m-%dT%H:%M:%S"))

            if delete_time + self.scrub_time > time.time():
                continue

            for loc in image['location_data']:
                if loc['status'] != 'pending_delete':
                    continue

                if self.metadata_encryption_key:
                    uri = crypt.urlsafe_encrypt(self.metadata_encryption_key,
                                                loc['url'], 64)
                else:
                    uri = loc['url']

                ret.append((image['id'], loc['id'], uri))
        return ret

    def has_image(self, image_id):
        """Returns whether the queue contains an image or not.

        :param image_id: The opaque image identifier

        :returns: a boolean value to inform including or not
        """
        try:
            image = self.registry.get_image(image_id)
            return image['status'] == 'pending_delete'
        except exception.NotFound:
            return False


_db_queue = None


def get_scrub_queue():
    global _db_queue
    if not _db_queue:
        _db_queue = ScrubDBQueue()
    return _db_queue


class Daemon(object):
    def __init__(self, wakeup_time=300, threads=100):
        LOG.info(_LI("Starting Daemon: wakeup_time=%(wakeup_time)s "
                     "threads=%(threads)s"),
                 {'wakeup_time': wakeup_time, 'threads': threads})
        self.wakeup_time = wakeup_time
        self.event = eventlet.event.Event()
        # This pool is used for periodic instantiation of scrubber
        self.daemon_pool = eventlet.greenpool.GreenPool(threads)

    def start(self, application):
        self._run(application)

    def wait(self):
        try:
            self.event.wait()
        except KeyboardInterrupt:
            msg = _LI("Daemon Shutdown on KeyboardInterrupt")
            LOG.info(msg)

    def _run(self, application):
        LOG.debug("Running application")
        self.daemon_pool.spawn_n(application.run, self.event)
        eventlet.spawn_after(self.wakeup_time, self._run, application)
        LOG.debug("Next run scheduled in %s seconds", self.wakeup_time)


class Scrubber(object):
    def __init__(self, store_api):
        LOG.info(_LI("Initializing scrubber with configuration: %s"),
                 six.text_type({'registry_host': CONF.registry_host,
                                'registry_port': CONF.registry_port}))

        self.store_api = store_api

        registry.configure_registry_client()
        registry.configure_registry_admin_creds()

        # Here we create a request context with credentials to support
        # delayed delete when using multi-tenant backend storage
        admin_user = CONF.admin_user
        admin_tenant = CONF.admin_tenant_name

        if CONF.send_identity_headers:
            # When registry is operating in trusted-auth mode
            roles = [CONF.admin_role]
            self.admin_context = context.RequestContext(user=admin_user,
                                                        tenant=admin_tenant,
                                                        auth_token=None,
                                                        roles=roles)
            self.registry = registry.get_registry_client(self.admin_context)
        else:
            ctxt = context.RequestContext()
            self.registry = registry.get_registry_client(ctxt)
            auth_token = self.registry.auth_token
            self.admin_context = context.RequestContext(user=admin_user,
                                                        tenant=admin_tenant,
                                                        auth_token=auth_token)

        self.db_queue = get_scrub_queue()
        self.pool = eventlet.greenpool.GreenPool(CONF.scrub_pool_size)

    def _get_delete_jobs(self):
        try:
            records = self.db_queue.get_all_locations()
        except Exception as err:
            LOG.error(_LE("Can not get scrub jobs from queue: %s") %
                      encodeutils.exception_to_unicode(err))
            return {}

        delete_jobs = {}
        for image_id, loc_id, loc_uri in records:
            if image_id not in delete_jobs:
                delete_jobs[image_id] = []
            delete_jobs[image_id].append((image_id, loc_id, loc_uri))
        return delete_jobs

    def run(self, event=None):
        delete_jobs = self._get_delete_jobs()

        if delete_jobs:
            list(self.pool.starmap(self._scrub_image, delete_jobs.items()))

    def _scrub_image(self, image_id, delete_jobs):
        if len(delete_jobs) == 0:
            return

        LOG.info(_LI("Scrubbing image %(id)s from %(count)d locations."),
                 {'id': image_id, 'count': len(delete_jobs)})

        success = True
        for img_id, loc_id, uri in delete_jobs:
            try:
                self._delete_image_location_from_backend(img_id, loc_id, uri)
            except Exception:
                success = False

        if success:
            image = self.registry.get_image(image_id)
            if image['status'] == 'pending_delete':
                self.registry.update_image(image_id, {'status': 'deleted'})
            LOG.info(_LI("Image %s has been scrubbed successfully"), image_id)
        else:
            LOG.warn(_LW("One or more image locations couldn't be scrubbed "
                         "from backend. Leaving image '%s' in 'pending_delete'"
                         " status") % image_id)

    def _delete_image_location_from_backend(self, image_id, loc_id, uri):
        if CONF.metadata_encryption_key:
            uri = crypt.urlsafe_decrypt(CONF.metadata_encryption_key, uri)
        try:
            LOG.debug("Scrubbing image %s from a location.", image_id)
            try:
                self.store_api.delete_from_backend(uri, self.admin_context)
            except store_exceptions.NotFound:
                LOG.info(_LI("Image location for image '%s' not found in "
                             "backend; Marking image location deleted in "
                             "db."), image_id)

            if loc_id != '-':
                db_api.get_api().image_location_delete(self.admin_context,
                                                       image_id,
                                                       int(loc_id),
                                                       'deleted')
            LOG.info(_LI("Image %s is scrubbed from a location."), image_id)
        except Exception as e:
            LOG.error(_LE("Unable to scrub image %(id)s from a location. "
                          "Reason: %(exc)s ") %
                      {'id': image_id,
                       'exc': encodeutils.exception_to_unicode(e)})
            raise



# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import glance.common.exception as exc
import glance.common.glare.definitions as definitions
from glance.glare.domain import proxy
from glance.i18n import _


class ArtifactProxy(proxy.Artifact):
    def __init__(self, artifact, repo):
        super(ArtifactProxy, self).__init__(artifact)
        self.artifact = artifact
        self.repo = repo

    def set_type_specific_property(self, prop_name, value):
        if prop_name not in self.metadata.attributes.dependencies:
            return super(ArtifactProxy, self).set_type_specific_property(
                prop_name, value)
        # for every dependency have to transfer dep_id into a dependency itself
        if value is None:
            setattr(self.artifact, prop_name, None)
        else:
            if not isinstance(value, list):
                setattr(self.artifact, prop_name,
                        self._fetch_dependency(value))
            else:
                setattr(self.artifact, prop_name,
                        [self._fetch_dependency(dep_id) for dep_id in value])

    def _fetch_dependency(self, dep_id):
        # check for circular dependency id -> id
        if self.id == dep_id:
            raise exc.ArtifactCircularDependency()
        art = self.repo.get(artifact_id=dep_id)

        # repo returns a proxy of some level.
        # Need to find the base declarative artifact
        while not isinstance(art, definitions.ArtifactType):
            art = art.base
        return art


class ArtifactRepo(proxy.ArtifactRepo):
    def __init__(self, repo, plugins,
                 item_proxy_class=None, item_proxy_kwargs=None):
        self.plugins = plugins
        super(ArtifactRepo, self).__init__(repo,
                                           item_proxy_class=ArtifactProxy,
                                           item_proxy_kwargs={'repo': self})

    def _check_dep_state(self, dep, state):
        """Raises an exception if dependency 'dep' is not in state 'state'"""
        if dep.state != state:
            raise exc.Invalid(_(
                "Not all dependencies are in '%s' state") % state)

    def publish(self, artifact, *args, **kwargs):
        """
        Creates transitive dependencies,
        checks that all dependencies are in active state and
        transfers artifact from creating to active state
        """
        # make sure that all required dependencies exist
        artifact.__pre_publish__(*args, **kwargs)
        # make sure that all dependencies are active
        for param in artifact.metadata.attributes.dependencies:
            dependency = getattr(artifact, param)
            if isinstance(dependency, list):
                for dep in dependency:
                    self._check_dep_state(dep, 'active')
            elif dependency:
                self._check_dep_state(dependency, 'active')
        # as state is changed on db save, have to retrieve the freshly changed
        # artifact (the one passed into the func will have old state value)
        artifact = self.base.publish(self.helper.unproxy(artifact))

        return self.helper.proxy(artifact)

    def remove(self, artifact):
        """
        Checks that artifact has no dependencies and removes it.
        Otherwise an exception is raised
        """
        for param in artifact.metadata.attributes.dependencies:
            if getattr(artifact, param):
                raise exc.Invalid(_(
                    "Dependency property '%s' has to be deleted first") %
                    param)
        return self.base.remove(self.helper.unproxy(artifact))


class ArtifactFactory(proxy.ArtifactFactory):
    def __init__(self, base, klass, repo):
        self.klass = klass
        self.repo = repo
        super(ArtifactFactory, self).__init__(
            base, artifact_proxy_class=ArtifactProxy,
            artifact_proxy_kwargs={'repo': self.repo})

    def new_artifact(self, *args, **kwargs):
        """
        Creates an artifact without dependencies first
        and then adds them to the newly created artifact
        """
        # filter dependencies
        no_deps = {p: kwargs[p] for p in kwargs
                   if p not in self.klass.metadata.attributes.dependencies}
        deps = {p: kwargs[p] for p in kwargs
                if p in self.klass.metadata.attributes.dependencies}
        artifact = super(ArtifactFactory, self).new_artifact(*args, **no_deps)
        # now set dependencies
        for dep_param, dep_value in deps.items():
            setattr(artifact, dep_param, dep_value)
        return artifact



# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

import six
from six.moves import range

from glance.common import exception as exc
from glance import context
import glance.glare as ga
import glance.tests.functional.db as db_tests
from glance.tests import utils as test_utils


UUID1, UUID2 = ('80cc6551-9db4-42aa-bb58-51c48757f285',
                'f89c675a-e01c-436c-a384-7d2e784fb2d9')
TYPE_NAME = u'TestArtifactType'
TYPE_VERSION = u'1.0.0'


class ArtifactsTestDriver(test_utils.BaseTestCase):
    def setUp(self):
        super(ArtifactsTestDriver, self).setUp()
        context_cls = context.RequestContext
        self.adm_context = context_cls(is_admin=True,
                                       auth_token='user:user:admin',
                                       tenant='admin-tenant')
        self.context = context_cls(is_admin=False,
                                   auth_token='user:user:user',
                                   tenant='test-tenant')
        self.db_api = db_tests.get_db(self.config)
        db_tests.reset_db(self.db_api)
        self.create_test_artifacts()

    def create_test_artifacts(self):
        dependency = {'2->1': [UUID1]}
        self.db_api.artifact_create(self.adm_context,
                                    get_fixture(id=UUID1,
                                                name="TestArtifact1",
                                                visibility="public"),
                                    TYPE_NAME,
                                    TYPE_VERSION)
        self.db_api.artifact_create(self.adm_context,
                                    get_fixture(id=UUID2,
                                                name="TestArtifact2",
                                                visibility="public",
                                                dependencies=dependency),
                                    TYPE_NAME,
                                    TYPE_VERSION)
        self.art1 = self.db_api.artifact_get(self.context, UUID1, TYPE_NAME,
                                             TYPE_VERSION)
        self.art2 = self.db_api.artifact_get(self.context, UUID2, TYPE_NAME,
                                             TYPE_VERSION)


class ArtifactTests(object):
    def test_artifact_create(self):
        artifact = get_fixture()
        created = self.db_api.artifact_create(self.context, artifact,
                                              TYPE_NAME, TYPE_VERSION)
        self.assertIsNotNone(created)
        self.assertEqual(artifact['name'], created['name'])
        self.assertEqual(artifact['type_name'], created['type_name'])
        self.assertEqual(artifact['type_version'], created['type_version'])

    def test_artifact_create_none_valued_props(self):
        artifact = get_fixture()
        artifact['properties']['lylyly'] = dict(value=None, type='int')
        artifact['properties']['hihihi'] = dict(value=5, type='int')
        created = self.db_api.artifact_create(self.context, artifact,
                                              TYPE_NAME, TYPE_VERSION)
        self.assertIsNotNone(created)
        self.assertIn('hihihi', created['properties'])
        self.assertNotIn('lylyly', created['properties'])

    def test_artifact_update(self):
        fixture = {'name': 'UpdatedName'}
        updated = self.db_api.artifact_update(self.context, fixture, UUID1,
                                              TYPE_NAME, TYPE_VERSION)
        self.assertIsNotNone(updated)
        self.assertEqual('UpdatedName', updated['name'])
        self.assertNotEqual(updated['created_at'], updated['updated_at'])

    def test_artifact_create_same_version_different_users(self):
        tenant1 = str(uuid.uuid4())
        tenant2 = str(uuid.uuid4())
        ctx1 = context.RequestContext(is_admin=False, tenant=tenant1)
        ctx2 = context.RequestContext(is_admin=False, tenant=tenant2)
        artifact1 = get_fixture(owner=tenant1)
        artifact2 = get_fixture(owner=tenant2)
        self.db_api.artifact_create(ctx1, artifact1,
                                    TYPE_NAME, TYPE_VERSION)

        self.assertIsNotNone(
            self.db_api.artifact_create(ctx2, artifact2,
                                        TYPE_NAME, TYPE_VERSION))

    def test_artifact_create_same_version_deleted(self):
        artifact1 = get_fixture()
        artifact2 = get_fixture(state='deleted')
        artifact3 = get_fixture(state='deleted')
        self.db_api.artifact_create(self.context, artifact1,
                                    TYPE_NAME, TYPE_VERSION)

        self.assertIsNotNone(
            self.db_api.artifact_create(self.context, artifact2,
                                        TYPE_NAME, TYPE_VERSION))
        self.assertIsNotNone(
            self.db_api.artifact_create(self.context, artifact3,
                                        TYPE_NAME, TYPE_VERSION))

    def test_artifact_get(self):
        res = self.db_api.artifact_get(self.context, UUID1,
                                       TYPE_NAME, TYPE_VERSION)
        self.assertEqual('TestArtifact1', res['name'])
        self.assertEqual('TestArtifactType', res['type_name'])
        self.assertEqual('1.0.0', res['type_version'])
        self.assertEqual('10.0.3-alpha+some-date', res['version'])
        self.assertEqual('creating', res['state'])
        self.assertEqual('test-tenant', res['owner'])

    def test_artifact_get_owned(self):
        tenant1 = str(uuid.uuid4())
        tenant2 = str(uuid.uuid4())
        ctx1 = context.RequestContext(is_admin=False, tenant=tenant1)
        ctx2 = context.RequestContext(is_admin=False, tenant=tenant2)

        artifact = get_fixture(owner=tenant1)
        created = self.db_api.artifact_create(ctx1, artifact,
                                              TYPE_NAME, TYPE_VERSION)
        self.assertIsNotNone(self.db_api.artifact_get(ctx1, created['id'],
                                                      TYPE_NAME, TYPE_VERSION))
        self.assertRaises(exc.ArtifactForbidden, self.db_api.artifact_get,
                          ctx2, created['id'], TYPE_NAME, TYPE_VERSION)

    def test_artifact_get_public(self):
        tenant1 = str(uuid.uuid4())
        tenant2 = str(uuid.uuid4())
        ctx1 = context.RequestContext(is_admin=False, tenant=tenant1)
        ctx2 = context.RequestContext(is_admin=False, tenant=tenant2)

        artifact = get_fixture(owner=tenant1, visibility='public')
        created = self.db_api.artifact_create(ctx1, artifact,
                                              TYPE_NAME, TYPE_VERSION)
        self.assertIsNotNone(self.db_api.artifact_get(ctx1, created['id'],
                                                      TYPE_NAME, TYPE_VERSION))
        self.assertIsNotNone(self.db_api.artifact_get(ctx2, created['id'],
                                                      TYPE_NAME, TYPE_VERSION))

    def test_artifact_update_state(self):
        res = self.db_api.artifact_update(self.context, {'state': 'active'},
                                          UUID1, TYPE_NAME, TYPE_VERSION)
        self.assertEqual('active', res['state'])

        self.assertRaises(exc.InvalidArtifactStateTransition,
                          self.db_api.artifact_update, self.context,
                          {'state': 'creating'}, UUID1,
                          TYPE_NAME, TYPE_VERSION)

        res = self.db_api.artifact_update(self.context,
                                          {'state': 'deactivated'}, UUID1,
                                          TYPE_NAME, TYPE_VERSION)
        self.assertEqual('deactivated', res['state'])
        res = self.db_api.artifact_update(self.context, {'state': 'active'},
                                          UUID1, TYPE_NAME, TYPE_VERSION)
        self.assertEqual('active', res['state'])
        res = self.db_api.artifact_update(self.context, {'state': 'deleted'},
                                          UUID1, TYPE_NAME, TYPE_VERSION)
        self.assertEqual('deleted', res['state'])

        self.assertRaises(exc.InvalidArtifactStateTransition,
                          self.db_api.artifact_update, self.context,
                          {'state': 'active'}, UUID1,
                          TYPE_NAME, TYPE_VERSION)
        self.assertRaises(exc.InvalidArtifactStateTransition,
                          self.db_api.artifact_update, self.context,
                          {'state': 'deactivated'}, UUID1,
                          TYPE_NAME, TYPE_VERSION)
        self.assertRaises(exc.InvalidArtifactStateTransition,
                          self.db_api.artifact_update, self.context,
                          {'state': 'creating'}, UUID1,
                          TYPE_NAME, TYPE_VERSION)

    def test_artifact_update_tags(self):
        res = self.db_api.artifact_update(self.context,
                                          {'tags': ['gagaga', 'lalala']},
                                          UUID1, TYPE_NAME, TYPE_VERSION)
        self.assertEqual(set(['gagaga', 'lalala']), set(res['tags']))

    def test_artifact_update_properties(self):
        new_properties = {'properties': {
            'propname1': {
                'type': 'string',
                'value': 'qeqeqe'},
            'propname2': {
                'type': 'int',
                'value': 6},
            'propname3': {
                'type': 'int',
                'value': '5'},
            'proparray': {
                'type': 'string',
                'value': 'notarray'
            }}
        }
        res = self.db_api.artifact_update(self.context,
                                          new_properties,
                                          UUID1, TYPE_NAME, TYPE_VERSION)
        bd_properties = res['properties']
        self.assertEqual(4, len(bd_properties))

        for prop in bd_properties:
            self.assertIn(prop, new_properties['properties'])

    def test_artifact_update_blobs(self):
        new_blobs = {'blobs': {
            'blob1': [{
                'size': 2600000,
                'checksum': 'abc',
                'item_key': 'some',
                'locations': [
                    {'value': 'URL11',
                     'status': 'active'},
                    {'value': 'URL12',
                     'status': 'active'}]
            }, {
                'size': 200000,
                'checksum': 'abc',
                'item_key': 'some',
                'locations': [
                    {'value': 'newURL21',
                     'status': 'active'},
                    {'value': 'URL22',
                     'status': 'passive'}]
            }
            ],
            'blob2': [{
                'size': 120000,
                'checksum': 'abc',
                'item_key': 'some',
                'locations': [
                    {'value': 'URL21',
                     'status': 'active'},
                    {'value': 'URL22',
                     'status': 'active'}]
            }, {
                'size': 300000,
                'checksum': 'abc',
                'item_key': 'some',
                'locations': [
                    {'value': 'URL21',
                     'status': 'active'},
                    {'value': 'bl1URL2',
                     'status': 'passive'}]
            }
            ]
        }

        }
        res = self.db_api.artifact_update(self.context,
                                          new_blobs,
                                          UUID1, TYPE_NAME, TYPE_VERSION)
        bd_blobs = res['blobs']
        self.assertEqual(2, len(bd_blobs))
        for blob in bd_blobs:
            self.assertIn(blob, new_blobs['blobs'])

    def test_artifact_create_with_dependency(self):
        dependencies = {"new->2": [UUID2]}
        artifact = get_fixture(dependencies=dependencies)
        res = self.db_api.artifact_create(self.context, artifact,
                                          TYPE_NAME, TYPE_VERSION)
        self.assertIsNotNone(res)

        created = self.db_api.artifact_get(
            self.context, res['id'], TYPE_NAME, TYPE_VERSION,
            show_level=ga.Showlevel.DIRECT)
        bd_dependencies = created['dependencies']
        self.assertEqual(1, len(bd_dependencies))
        # now try to update artifact with the same dependency
        new_dependencies = {"dependencies": {"new->2": [UUID2],
                                             "new->3": [UUID2]}}
        res = self.db_api.artifact_update(self.context,
                                          new_dependencies,
                                          UUID1, TYPE_NAME, TYPE_VERSION)
        retrieved = self.db_api.artifact_get(
            self.context, res['id'],
            TYPE_NAME, TYPE_VERSION, show_level=ga.Showlevel.DIRECT)
        self.assertEqual(2, len(retrieved["dependencies"]))

    def test_artifact_create_transitive_dependencies(self):
        dependencies = {"new->2": [UUID2]}
        artifact = get_fixture(dependencies=dependencies, id='new')
        res = self.db_api.artifact_create(self.context, artifact,
                                          TYPE_NAME, TYPE_VERSION)
        self.assertIsNotNone(res)

        created = self.db_api.artifact_get(
            self.context, res['id'], TYPE_NAME, TYPE_VERSION,
            show_level=ga.Showlevel.DIRECT)
        bd_dependencies = created['dependencies']
        self.assertEqual(1, len(bd_dependencies))

        res = self.db_api.artifact_publish(
            self.context,
            res['id'], TYPE_NAME, TYPE_VERSION
        )

        res = self.db_api.artifact_get(
            self.context, res['id'], TYPE_NAME, TYPE_VERSION,
            show_level=ga.Showlevel.TRANSITIVE)
        self.assertIsNotNone(res.pop('created_at'))
        self.assertIsNotNone(res.pop('updated_at'))

        # NOTE(mfedosin): tags is a set, so we have to check it separately
        tags = res.pop('tags', None)
        self.assertIsNotNone(tags)
        self.assertEqual(set(['gugugu', 'lalala']), set(tags))

        tags = res['dependencies']['new->2'][0].pop('tags', None)
        self.assertIsNotNone(tags)
        self.assertEqual(set(['gugugu', 'lalala']), set(tags))

        tags = (res['dependencies']['new->2'][0]['dependencies']['2->1'][0].
                pop('tags', None))
        self.assertIsNotNone(tags)
        self.assertEqual(set(['gugugu', 'lalala']), set(tags))

        expected = {
            'id': 'new',
            'name': u'SomeArtifact',
            'description': None,
            'type_name': TYPE_NAME,
            'type_version': TYPE_VERSION,
            'version': u'10.0.3-alpha+some-date',
            'visibility': u'private',
            'state': u'active',
            'owner': u'test-tenant',
            'published_at': None,
            'deleted_at': None,
            'properties': {
                'propname1': {
                    'type': 'string',
                    'value': 'tututu'},
                'propname2': {
                    'type': 'int',
                    'value': 5},
                'propname3': {
                    'type': 'string',
                    'value': 'vavava'},
                'proparray': {
                    'type': 'array',
                    'value': [
                        {'type': 'int',
                         'value': 6},
                        {'type': 'string',
                         'value': 'rerere'}
                    ]
                }
            },
            'blobs': {
                'blob1': [{
                    'size': 1600000,
                    'checksum': 'abc',
                    'item_key': 'some',
                    'locations': [
                        {'value': 'URL11',
                         'status': 'active'},
                        {'value': 'URL12',
                         'status': 'active'}]
                }, {
                    'size': 100000,
                    'checksum': 'abc',
                    'item_key': 'some',
                    'locations': [
                        {'value': 'URL21',
                         'status': 'active'},
                        {'value': 'URL22',
                         'status': 'active'}]
                }]
            },
            'dependencies': {
                'new->2': [
                    {
                        'id': UUID2,
                        'created_at': self.art2['created_at'],
                        'updated_at': self.art2['updated_at'],
                        'published_at': None,
                        'deleted_at': None,
                        'name': u'TestArtifact2',
                        'description': None,
                        'type_name': TYPE_NAME,
                        'type_version': TYPE_VERSION,
                        'version': u'10.0.3-alpha+some-date',
                        'visibility': 'public',
                        'state': u'creating',
                        'owner': u'test-tenant',
                        'properties': {
                            'propname1': {
                                'type': 'string',
                                'value': 'tututu'},
                            'propname2': {
                                'type': 'int',
                                'value': 5},
                            'propname3': {
                                'type': 'string',
                                'value': 'vavava'},
                            'proparray': {
                                'type': 'array',
                                'value': [
                                    {'type': 'int',
                                     'value': 6},
                                    {'type': 'string',
                                     'value': 'rerere'}
                                ]
                            }
                        },
                        'blobs': {
                            'blob1': [{
                                'size': 1600000,
                                'checksum': 'abc',
                                'item_key': 'some',
                                'locations': [
                                    {'value': 'URL11',
                                     'status': 'active'},
                                    {'value': 'URL12',
                                     'status': 'active'}]
                            }, {
                                'size': 100000,
                                'checksum': 'abc',
                                'item_key': 'some',
                                'locations': [
                                    {'value': 'URL21',
                                     'status': 'active'},
                                    {'value': 'URL22',
                                     'status': 'active'}]
                            }]
                        },
                        'dependencies': {
                            '2->1': [
                                {
                                    'id': UUID1,
                                    'created_at': self.art1['created_at'],
                                    'updated_at': self.art1['updated_at'],
                                    'published_at': None,
                                    'deleted_at': None,
                                    'dependencies': {},
                                    'name': u'TestArtifact1',
                                    'description': None,
                                    'type_name': TYPE_NAME,
                                    'type_version': TYPE_VERSION,
                                    'version': u'10.0.3-alpha+some-date',
                                    'visibility': 'public',
                                    'state': u'creating',
                                    'owner': u'test-tenant',
                                    'properties': {
                                        'propname1': {
                                            'type': 'string',
                                            'value': 'tututu'},
                                        'propname2': {
                                            'type': 'int',
                                            'value': 5},
                                        'propname3': {
                                            'type': 'string',
                                            'value': 'vavava'},
                                        'proparray': {
                                            'type': 'array',
                                            'value': [
                                                {'type': 'int',
                                                 'value': 6},
                                                {'type': 'string',
                                                 'value': 'rerere'}
                                            ]
                                        }
                                    },
                                    'blobs': {
                                        'blob1': [{
                                            'size': 1600000,
                                            'checksum': 'abc',
                                            'item_key': 'some',
                                            'locations': [
                                                {'value': 'URL11',
                                                 'status': 'active'},
                                                {'value': 'URL12',
                                                 'status': 'active'}]
                                        }, {
                                            'size': 100000,
                                            'checksum': 'abc',
                                            'item_key': 'some',
                                            'locations': [
                                                {'value': 'URL21',
                                                 'status': 'active'},
                                                {'value': 'URL22',
                                                 'status': 'active'}]
                                        }]
                                    }
                                }
                            ]
                        }
                    }
                ]
            }
        }
        self.assertIsNotNone(res['published_at'])
        published_at = res['published_at']
        expected['published_at'] = published_at
        for key, value in six.iteritems(expected):
            self.assertEqual(expected[key], res[key])

    def test_artifact_get_all(self):
        artifact = get_fixture(name='new_artifact')
        self.db_api.artifact_create(self.context, artifact,
                                    TYPE_NAME, TYPE_VERSION)
        artifacts = self.db_api.artifact_get_all(self.context)
        self.assertEqual(3, len(artifacts))

    def test_artifact_sort_order(self):
        arts = [get_fixture(version='1.2.3-alpha.4.df.00f'),
                get_fixture(version='1.2.2'),
                get_fixture(version='1.2.3+some-metadata'),
                get_fixture(version='1.2.4'),
                get_fixture(version='1.2.3-release.2'),
                get_fixture(version='1.2.3-release.1+metadata'),
                get_fixture(version='1.2.3-final'),
                get_fixture(version='1.2.3-alpha.14.df.00f')]
        for art in arts:
            self.db_api.artifact_create(self.context, art, TYPE_NAME,
                                        TYPE_VERSION)
        artifacts = self.db_api.artifact_get_all(self.context,
                                                 sort_keys=[('version',
                                                             None)],
                                                 sort_dirs=['asc'])

        expected_versions = [
            '1.2.2',
            '1.2.3-alpha.4.df.00f',
            '1.2.3-alpha.14.df.00f',
            '1.2.3-final',
            '1.2.3-release.1+metadata',
            '1.2.3-release.2',
            '1.2.3+some-metadata',
            '1.2.4']
        for i in range(len(expected_versions)):
            self.assertEqual(expected_versions[i], artifacts[i]['version'])

    def test_artifact_get_all_show_level(self):
        artifacts = self.db_api.artifact_get_all(self.context)
        self.assertEqual(2, len(artifacts))

        self.assertRaises(KeyError, lambda: artifacts[0]['properties'])

        artifacts = self.db_api.artifact_get_all(
            self.context, show_level=ga.Showlevel.BASIC)
        self.assertEqual(2, len(artifacts))
        self.assertEqual(4, len(artifacts[0]['properties']))

        self.assertRaises(exc.ArtifactUnsupportedShowLevel,
                          self.db_api.artifact_get_all, self.context,
                          show_level=ga.Showlevel.DIRECT)

    def test_artifact_get_all_tags(self):
        artifact = get_fixture(name='new_artifact',
                               tags=['qwerty', 'uiop'])
        self.db_api.artifact_create(self.context, artifact,
                                    TYPE_NAME, TYPE_VERSION)
        artifacts = self.db_api.artifact_get_all(self.context)
        self.assertEqual(3, len(artifacts))

        filters = {'tags': [{
            'value': 'notag',
        }]}
        artifacts = self.db_api.artifact_get_all(self.context, filters=filters)
        self.assertEqual(0, len(artifacts))

        filters = {'tags': [{
            'value': 'lalala',
        }]}
        artifacts = self.db_api.artifact_get_all(self.context, filters=filters)
        self.assertEqual(2, len(artifacts))
        for artifact in artifacts:
            self.assertIn(artifact['name'], ['TestArtifact1', 'TestArtifact2'])

    def test_artifact_get_all_properties(self):
        artifact = get_fixture(
            name='new_artifact',
            properties={
                'newprop2': {
                    'type': 'string',
                    'value': 'tututu'},
                'propname2': {
                    'type': 'int',
                    'value': 3},
                'propname3': {
                    'type': 'string',
                    'value': 'vavava'},
                'proptext': {
                    'type': 'text',
                    'value': 'bebebe' * 100},
                'proparray': {
                    'type': 'array',
                    'value': [
                        {'type': 'int',
                         'value': 17},
                        {'type': 'string',
                         'value': 'rerere'}
                    ]
                }})
        self.db_api.artifact_create(self.context, artifact,
                                    TYPE_NAME, TYPE_VERSION)

        filters = {'propname2': [{
            'value': 4,
            'operator': 'GT',
            'type': 'int'}]}
        artifacts = self.db_api.artifact_get_all(self.context, filters=filters)
        self.assertEqual(2, len(artifacts))
        for artifact in artifacts:
            self.assertIn(artifact['name'], ['TestArtifact1', 'TestArtifact2'])

        # position hasn't been set
        filters = {'proparray': [{
            'value': 6,
            'operator': 'LE',
            'type': 'int'}]}
        artifacts = self.db_api.artifact_get_all(self.context, filters=filters)
        self.assertEqual(0, len(artifacts))
        for artifact in artifacts:
            self.assertIn(artifact['name'], ['TestArtifact1', 'TestArtifact2'])

        # position has been set
        filters = {'proparray': [{
            'value': 6,
            'position': 0,
            'operator': 'LE',
            'type': 'int'}]}
        artifacts = self.db_api.artifact_get_all(self.context, filters=filters)
        self.assertEqual(2, len(artifacts))
        for artifact in artifacts:
            self.assertIn(artifact['name'], ['TestArtifact1', 'TestArtifact2'])

        filters = {'proparray': [{
            'value': 6,
            'operator': 'IN',
            'type': 'int'}]}
        artifacts = self.db_api.artifact_get_all(self.context, filters=filters)
        self.assertEqual(2, len(artifacts))
        for artifact in artifacts:
            self.assertIn(artifact['name'], ['TestArtifact1', 'TestArtifact2'])

        filters = {'name': [{'value': 'new_artifact'}]}
        artifacts = self.db_api.artifact_get_all(self.context,
                                                 filters=filters,
                                                 show_level=ga.Showlevel.BASIC)
        self.assertEqual(1, len(artifacts))
        artifact = artifacts[0]
        self.assertEqual('new_artifact', artifact['name'])
        for prop in artifact['properties'].keys():
            self.assertNotEqual('proptext', prop)

        filters = {'propname2': [{
            'value': 4,
            'operator': 'FOO',
            'type': 'int'}]}
        self.assertRaises(
            exc.ArtifactUnsupportedPropertyOperator,
            self.db_api.artifact_get_all, self.context, filters=filters)

    def test_artifact_delete(self):
        res = self.db_api.artifact_delete(self.context, UUID1,
                                          TYPE_NAME, TYPE_VERSION)
        self.assertEqual('TestArtifact1', res['name'])
        self.assertEqual('deleted', res['state'])
        self.assertIsNotNone(res['deleted_at'])

        artifacts = self.db_api.artifact_get_all(self.context)
        self.assertEqual(1, len(artifacts))

    def test_artifact_delete_property(self):

        new_properties = {'properties': {
            'proparray': {'value': [],
                          'type': 'array'}
        }
        }
        res = self.db_api.artifact_update(self.context,
                                          new_properties,
                                          UUID1, TYPE_NAME, TYPE_VERSION)
        bd_properties = res['properties']
        self.assertEqual(3, len(bd_properties))

        expected = {
            'propname1': {
                'type': 'string',
                'value': 'tututu'},
            'propname2': {
                'type': 'int',
                'value': 5},
            'propname3': {
                'type': 'string',
                'value': 'vavava'}
        }

        for prop in bd_properties:
            self.assertIn(prop, expected)

    def test_artifact_delete_blob(self):

        new_blobs = {'blobs': {
            'blob2': [{
                'size': 2600000,
                'checksum': 'abc',
                'item_key': 'some',
                'locations': [
                    {'value': 'URL11',
                     'status': 'active'},
                    {'value': 'URL12',
                     'status': 'active'}]
            }, {
                'size': 200000,
                'checksum': 'abc',
                'item_key': 'some',
                'locations': [
                    {'value': 'newURL21',
                     'status': 'active'},
                    {'value': 'URL22',
                     'status': 'passive'}]
            }
            ],
            'blob3': [{
                'size': 120000,
                'checksum': 'abc',
                'item_key': 'some',
                'locations': [
                    {'value': 'URL21',
                     'status': 'active'},
                    {'value': 'URL22',
                     'status': 'active'}]
            }, {
                'size': 300000,
                'checksum': 'abc',
                'item_key': 'some',
                'locations': [
                    {'value': 'URL21',
                     'status': 'active'},
                    {'value': 'bl1URL2',
                     'status': 'passive'}]
            }
            ]
        }

        }

        expected = {'blobs': {
            'blob1': [{
                'size': 1600000,
                'checksum': 'abc',
                'item_key': 'some',
                'locations': [
                    {'value': 'URL11',
                     'status': 'active'},
                    {'value': 'URL12',
                     'status': 'active'}]
            }, {
                'size': 100000,
                'checksum': 'abc',
                'item_key': 'some',
                'locations': [
                    {'value': 'URL21',
                     'status': 'active'},
                    {'value': 'URL22',
                     'status': 'active'}]
            }
            ],
            'blob2': [{
                'size': 2600000,
                'checksum': 'abc',
                'item_key': 'some',
                'locations': [
                    {'value': 'URL11',
                     'status': 'active'},
                    {'value': 'URL12',
                     'status': 'active'}]
            }, {
                'size': 200000,
                'checksum': 'abc',
                'item_key': 'some',
                'locations': [
                    {'value': 'newURL21',
                     'status': 'active'},
                    {'value': 'URL22',
                     'status': 'passive'}]
            }
            ],
            'blob3': [{
                'size': 120000,
                'checksum': 'abc',
                'item_key': 'some',
                'locations': [
                    {'value': 'URL21',
                     'status': 'active'},
                    {'value': 'URL22',
                     'status': 'active'}]
            }, {
                'size': 300000,
                'checksum': 'abc',
                'item_key': 'some',
                'locations': [
                    {'value': 'URL21',
                     'status': 'active'},
                    {'value': 'bl1URL2',
                     'status': 'passive'}]
            }
            ]
        }

        }

        res = self.db_api.artifact_update(self.context,
                                          new_blobs,
                                          UUID1, TYPE_NAME, TYPE_VERSION)
        bd_blobs = res['blobs']
        self.assertEqual(3, len(bd_blobs))
        for blob in bd_blobs:
            self.assertIn(blob, expected['blobs'])

        del_blobs = {'blobs': {
            'blob1': []}
        }

        res = self.db_api.artifact_update(self.context,
                                          del_blobs,
                                          UUID1, TYPE_NAME, TYPE_VERSION)
        bd_blobs = res['blobs']
        self.assertEqual(2, len(bd_blobs))

        for blob in bd_blobs:
            self.assertIn(blob, new_blobs['blobs'])


def get_fixture(**kwargs):
    artifact = {
        'name': u'SomeArtifact',
        'type_name': TYPE_NAME,
        'type_version': TYPE_VERSION,
        'version': u'10.0.3-alpha+some-date',
        'visibility': u'private',
        'state': u'creating',
        'owner': u'test-tenant',
        'tags': ['lalala', 'gugugu'],
        'properties': {
            'propname1': {
                'type': 'string',
                'value': 'tututu'},
            'propname2': {
                'type': 'int',
                'value': 5},
            'propname3': {
                'type': 'string',
                'value': 'vavava'},
            'proparray': {
                'type': 'array',
                'value': [
                    {'type': 'int',
                     'value': 6},
                    {'type': 'string',
                     'value': 'rerere'}
                ]
            }
        },
        'blobs': {
            'blob1': [{
                'size': 1600000,
                'checksum': 'abc',
                'item_key': 'some',
                'locations': [
                    {'value': 'URL11',
                     'status': 'active'},
                    {'value': 'URL12',
                     'status': 'active'}]
            }, {
                'size': 100000,
                'checksum': 'abc',
                'item_key': 'some',
                'locations': [
                    {'value': 'URL21',
                     'status': 'active'},
                    {'value': 'URL22',
                     'status': 'active'}]
            }
            ]
        }
    }

    artifact.update(kwargs)
    return artifact



# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import re

import semantic_version
import six

from glance.common import exception as exc
from glance.i18n import _


class AttributeDefinition(object):
    """A base class for the attribute definitions which may be added to
    declaratively defined artifact types
    """

    ALLOWED_TYPES = (object,)

    def __init__(self,
                 display_name=None,
                 description=None,
                 readonly=False,
                 mutable=True,
                 required=False,
                 default=None):
        """Initializes attribute definition

        :param display_name: Display name of the attribute
        :param description: Description of the attribute
        :param readonly: Flag indicating if the value of attribute may not be
        changed once an artifact is created
        :param mutable: Flag indicating if the value of attribute may not be
        changed once an artifact is published
        :param required: Flag indicating if the value of attribute is required
        :param default: default value of the attribute
        """
        self.name = None
        self.display_name = display_name
        self.description = description
        self.readonly = readonly
        self.required = required
        self.mutable = mutable
        self.default = default
        self._add_validator('type',
                            lambda v: isinstance(v, self.ALLOWED_TYPES),
                            _("Not a valid value type"))
        self._validate_default()

    def _set_name(self, value):
        self.name = value
        if self.display_name is None:
            self.display_name = value

    def _add_validator(self, name, func, message):
        if not hasattr(self, '_validators'):
            self._validators = []
            self._validators_index = {}
        pair = (func, message)
        self._validators.append(pair)
        self._validators_index[name] = pair

    def _get_validator(self, name):
        return self._validators_index.get(name)

    def _remove_validator(self, name):
        pair = self._validators_index.pop(name, None)
        if pair is not None:
            self._validators.remove(pair)

    def _check_definition(self):
        self._validate_default()

    def _validate_default(self):
        if self.default:
            try:
                self.validate(self.default, 'default')
            except exc.InvalidArtifactPropertyValue:
                raise exc.InvalidArtifactTypePropertyDefinition(
                    _("Default value is invalid"))

    def get_value(self, obj):
        return getattr(obj, self.name)

    def set_value(self, obj, value):
        return setattr(obj, self.name, value)

    def validate(self, value, name=None):
        if value is None:
            if self.required:
                raise exc.InvalidArtifactPropertyValue(
                    name=name or self.name,
                    val=value,
                    msg=_('Value is required'))
            else:
                return

        first_error = next((msg for v_func, msg in self._validators
                            if not v_func(value)), None)
        if first_error:
            raise exc.InvalidArtifactPropertyValue(name=name or self.name,
                                                   val=value,
                                                   msg=first_error)


class ListAttributeDefinition(AttributeDefinition):
    """A base class for Attribute definitions having List-semantics

    Is inherited by Array, ArtifactReferenceList and BinaryObjectList
    """
    ALLOWED_TYPES = (list,)
    ALLOWED_ITEM_TYPES = (AttributeDefinition, )

    def _check_item_type(self, item):
        if not isinstance(item, self.ALLOWED_ITEM_TYPES):
            raise exc.InvalidArtifactTypePropertyDefinition(
                _('Invalid item type specification'))
        if item.default is not None:
            raise exc.InvalidArtifactTypePropertyDefinition(
                _('List definitions may hot have defaults'))

    def __init__(self, item_type, min_size=0, max_size=None, unique=False,
                 **kwargs):

        super(ListAttributeDefinition, self).__init__(**kwargs)
        if isinstance(item_type, list):
            for it in item_type:
                self._check_item_type(it)

            # we need to copy the item_type collection
            self.item_type = item_type[:]

            if min_size != 0:
                raise exc.InvalidArtifactTypePropertyDefinition(
                    _("Cannot specify 'min_size' explicitly")
                )

            if max_size is not None:
                raise exc.InvalidArtifactTypePropertyDefinition(
                    _("Cannot specify 'max_size' explicitly")
                )

            # setting max_size and min_size to the length of item_type,
            # as tuple-semantic assumes that the number of elements is set
            # by the type spec
            min_size = max_size = len(item_type)
        else:
            self._check_item_type(item_type)
            self.item_type = item_type

        if min_size:
            self.min_size(min_size)

        if max_size:
            self.max_size(max_size)

        if unique:
            self.unique()

    def min_size(self, value):
        self._min_size = value
        if value is not None:
            self._add_validator('min_size',
                                lambda v: len(v) >= self._min_size,
                                _('List size is less than minimum'))
        else:
            self._remove_validator('min_size')

    def max_size(self, value):
        self._max_size = value
        if value is not None:
            self._add_validator('max_size',
                                lambda v: len(v) <= self._max_size,
                                _('List size is greater than maximum'))
        else:
            self._remove_validator('max_size')

    def unique(self, value=True):
        self._unique = value
        if value:
            def _unique(items):
                seen = set()
                for item in items:
                    if item in seen:
                        return False
                    seen.add(item)
                return True
            self._add_validator('unique',
                                _unique, _('Items have to be unique'))
        else:
            self._remove_validator('unique')

    def _set_name(self, value):
        super(ListAttributeDefinition, self)._set_name(value)
        if isinstance(self.item_type, list):
            for i, item in enumerate(self.item_type):
                item._set_name("%s[%i]" % (value, i))
        else:
            self.item_type._set_name("%s[*]" % value)

    def validate(self, value, name=None):
        super(ListAttributeDefinition, self).validate(value, name)
        if value is not None:
            for i, item in enumerate(value):
                self._validate_item_at(item, i)

    def get_item_definition_at_index(self, index):
        if isinstance(self.item_type, list):
            if index < len(self.item_type):
                return self.item_type[index]
            else:
                return None
        return self.item_type

    def _validate_item_at(self, item, index):
        item_type = self.get_item_definition_at_index(index)
        # set name if none has been given to the list element at given index
        if (isinstance(self.item_type, list) and item_type and
                not item_type.name):
            item_type.name = "%s[%i]" % (self.name, index)
        if item_type:
            item_type.validate(item)


class DictAttributeDefinition(AttributeDefinition):
    """A base class for Attribute definitions having Map-semantics

    Is inherited by Dict
    """
    ALLOWED_TYPES = (dict,)
    ALLOWED_PROPERTY_TYPES = (AttributeDefinition,)

    def _check_prop(self, key, item):
        if (not isinstance(item, self.ALLOWED_PROPERTY_TYPES) or
                (key is not None and not isinstance(key, six.string_types))):
            raise exc.InvalidArtifactTypePropertyDefinition(
                _('Invalid dict property type specification'))

    @staticmethod
    def _validate_key(key):
        if not isinstance(key, six.string_types):
            raise exc.InvalidArtifactPropertyValue(
                _('Invalid dict property type'))

    def __init__(self, properties, min_properties=0, max_properties=0,
                 **kwargs):
        super(DictAttributeDefinition, self).__init__(**kwargs)
        if isinstance(properties, dict):
            for key, value in six.iteritems(properties):
                self._check_prop(key, value)
            # copy the properties dict
            self.properties = properties.copy()

            self._add_validator('keys',
                                lambda v: set(v.keys()) <= set(
                                    self.properties.keys()),
                                _('Dictionary contains unexpected key(s)'))
        else:
            self._check_prop(None, properties)
            self.properties = properties

        if min_properties:
            self.min_properties(min_properties)

        if max_properties:
            self.max_properties(max_properties)

    def min_properties(self, value):
        self._min_properties = value
        if value is not None:
            self._add_validator('min_properties',
                                lambda v: len(v) >= self._min_properties,
                                _('Dictionary size is less than '
                                  'minimum'))
        else:
            self._remove_validator('min_properties')

    def max_properties(self, value):
        self._max_properties = value
        if value is not None:
            self._add_validator('max_properties',
                                lambda v: len(v) <= self._max_properties,
                                _('Dictionary size is '
                                  'greater than maximum'))
        else:
            self._remove_validator('max_properties')

    def _set_name(self, value):
        super(DictAttributeDefinition, self)._set_name(value)
        if isinstance(self.properties, dict):
            for k, v in six.iteritems(self.properties):
                v._set_name(value)
        else:
            self.properties._set_name(value)

    def validate(self, value, name=None):
        super(DictAttributeDefinition, self).validate(value, name)
        if value is not None:
            for k, v in six.iteritems(value):
                self._validate_item_with_key(v, k)

    def _validate_item_with_key(self, value, key):
        self._validate_key(key)
        if isinstance(self.properties, dict):
            prop_def = self.properties.get(key)
            if prop_def is not None:
                name = "%s[%s]" % (prop_def.name, key)
                prop_def.validate(value, name=name)
        else:
            name = "%s[%s]" % (self.properties.name, key)
            self.properties.validate(value, name=name)

    def get_prop_definition_at_key(self, key):
        if isinstance(self.properties, dict):
            return self.properties.get(key)
        else:
            return self.properties


class PropertyDefinition(AttributeDefinition):
    """A base class for Attributes defining generic or type-specific metadata
    properties
    """
    DB_TYPE = None

    def __init__(self,
                 internal=False,
                 allowed_values=None,
                 validators=None,
                 **kwargs):
        """Defines a metadata property

        :param internal: a flag indicating that the property is internal, i.e.
        not returned to client
        :param allowed_values: specifies a list of values allowed for the
        property
        :param validators: specifies a list of custom validators for the
        property
        """
        super(PropertyDefinition, self).__init__(**kwargs)
        self.internal = internal
        self._allowed_values = None

        if validators is not None:
            try:
                for i, (f, m) in enumerate(validators):
                    self._add_validator("custom_%i" % i, f, m)
            except ValueError:
                raise exc.InvalidArtifactTypePropertyDefinition(
                    _("Custom validators list should contain tuples "
                      "'(function, message)'"))

        if allowed_values is not None:
            # copy the allowed_values, as this is going to create a
            # closure, and we need to make sure that external modification of
            # this list does not affect the created validator
            self.allowed_values(allowed_values)
        self._check_definition()

    def _validate_allowed_values(self):
        if self._allowed_values:
            try:
                for allowed_value in self._allowed_values:
                    self.validate(allowed_value, 'allowed_value')
            except exc.InvalidArtifactPropertyValue:
                raise exc.InvalidArtifactTypePropertyDefinition(
                    _("Allowed values %s are invalid under given validators") %
                    self._allowed_values)

    def allowed_values(self, values):
        self._allowed_values = values[:]
        if values is not None:
            self._add_validator('allowed', lambda v: v in self._allowed_values,
                                _("Is not allowed value"))
        else:
            self._remove_validator('allowed')
        self._check_definition()

    def _check_definition(self):
        self._validate_allowed_values()
        super(PropertyDefinition, self)._check_definition()


class RelationDefinition(AttributeDefinition):
    """A base class for Attributes defining cross-artifact relations"""
    def __init__(self, internal=False, **kwargs):
        self.internal = internal
        kwargs.setdefault('mutable', False)
        # if mutable=True has been passed -> raise an exception
        if kwargs['mutable'] is True:
            raise exc.InvalidArtifactTypePropertyDefinition(
                _("Dependency relations cannot be mutable"))
        super(RelationDefinition, self).__init__(**kwargs)


class BlobDefinition(AttributeDefinition):
    """A base class for Attributes defining binary objects"""
    pass


class ArtifactTypeMetaclass(type):
    """A metaclass to build  Artifact Types. Not intended to be used directly

    Use `get_declarative_base` to get the base class instead
    """
    def __init__(cls, class_name, bases, attributes):
        if '_declarative_artifact_type' not in cls.__dict__:
            _build_declarative_meta(cls)
        super(ArtifactTypeMetaclass, cls).__init__(class_name, bases,
                                                   attributes)


class ArtifactPropertyDescriptor(object):
    """A descriptor object for working with artifact attributes"""

    def __init__(self, prop, collection_wrapper_class=None):
        self.prop = prop
        self.collection_wrapper_class = collection_wrapper_class

    def __get__(self, instance, owner):
        if instance is None:
            # accessed via owner class
            return self.prop
        else:
            v = getattr(instance, '_' + self.prop.name, None)
            if v is None and self.prop.default is not None:
                v = copy.copy(self.prop.default)
                self.__set__(instance, v, ignore_mutability=True)
                return self.__get__(instance, owner)
            else:
                if v is not None and self.collection_wrapper_class:
                    if self.prop.readonly:
                        readonly = True
                    elif (not self.prop.mutable and
                          hasattr(instance, '__is_mutable__') and
                          not hasattr(instance,
                                      '__suspend_mutability_checks__')):

                        readonly = not instance.__is_mutable__()
                    else:
                        readonly = False
                    if readonly:
                        v = v.__make_immutable__()
                return v

    def __set__(self, instance, value, ignore_mutability=False):
        if instance:
            if self.prop.readonly:
                if hasattr(instance, '_' + self.prop.name):
                    raise exc.InvalidArtifactPropertyValue(
                        _('Attempt to set readonly property'))
            if not self.prop.mutable:
                if (hasattr(instance, '__is_mutable__') and
                        not hasattr(instance,
                                    '__suspend_mutability_checks__')):
                    mutable = instance.__is_mutable__() or ignore_mutability
                    if not mutable:
                        raise exc.InvalidArtifactPropertyValue(
                            _('Attempt to set value of immutable property'))
            if value is not None and self.collection_wrapper_class:
                value = self.collection_wrapper_class(value)
                value.property = self.prop
            self.prop.validate(value)
            setattr(instance, '_' + self.prop.name, value)


class ArtifactAttributes(object):
    """A container class storing description of Artifact Type attributes"""
    def __init__(self):
        self.properties = {}
        self.dependencies = {}
        self.blobs = {}
        self.all = {}

    @property
    def default_dependency(self):
        """Returns the default dependency relation for an artifact type"""
        if len(self.dependencies) == 1:
            return list(self.dependencies.values())[0]

    @property
    def default_blob(self):
        """Returns the default blob object for an artifact type"""
        if len(self.blobs) == 1:
            return list(self.blobs.values())[0]

    @property
    def default_properties_dict(self):
        """Returns a default properties dict for an artifact type"""
        dict_props = [v for v in self.properties.values() if
                      isinstance(v, DictAttributeDefinition)]
        if len(dict_props) == 1:
            return dict_props[0]

    @property
    def tags(self):
        """Returns tags property for an artifact type"""
        return self.properties.get('tags')

    def add(self, attribute):
        self.all[attribute.name] = attribute
        if isinstance(attribute, PropertyDefinition):
            self.properties[attribute.name] = attribute
        elif isinstance(attribute, BlobDefinition):
            self.blobs[attribute.name] = attribute
        elif isinstance(attribute, RelationDefinition):
            self.dependencies[attribute.name] = attribute


class ArtifactTypeMetadata(object):
    """A container to store the meta-information about an artifact type"""

    def __init__(self, type_name, type_display_name, type_version,
                 type_description, endpoint):
        """Initializes the Artifact Type metadata

        :param type_name: name of the artifact type
        :param type_display_name: display name of the artifact type
        :param type_version:  version of the artifact type
        :param type_description: description of the artifact type
        :param endpoint: REST API URI suffix to call the artifacts of this type
        """

        self.attributes = ArtifactAttributes()

        # These are going to be defined by third-party plugin
        # developers, so we need to do some validations on these values and
        # raise InvalidArtifactTypeDefinition if they are violated
        self.type_name = type_name
        self.type_display_name = type_display_name or type_name
        self.type_version = type_version or '1.0'
        self.type_description = type_description
        self.endpoint = endpoint or type_name.lower()

        self._validate_string(self.type_name, 'Type name', min_length=1,
                              max_length=255)
        self._validate_string(self.type_display_name, 'Type display name',
                              max_length=255)
        self._validate_string(self.type_description, 'Type description')
        self._validate_string(self.endpoint, 'endpoint', min_length=1)
        try:
            semantic_version.Version(self.type_version, partial=True)
        except ValueError:
            raise exc.InvalidArtifactTypeDefinition(
                message=_("Type version has to be a valid semver string"))

    @staticmethod
    def _validate_string(value, name, min_length=0, max_length=None,
                         pattern=None):
        if value is None:
            if min_length > 0:
                raise exc.InvalidArtifactTypeDefinition(
                    message=_("%(attribute)s is required"), attribute=name)
            else:
                return
        if not isinstance(value, six.string_types):
            raise exc.InvalidArtifactTypeDefinition(
                message=_("%(attribute)s have to be string"), attribute=name)
        if max_length and len(value) > max_length:
            raise exc.InvalidArtifactTypeDefinition(
                message=_("%(attribute)s may not be longer than %(length)i"),
                attribute=name, length=max_length)
        if min_length and len(value) < min_length:
            raise exc.InvalidArtifactTypeDefinition(
                message=_("%(attribute)s may not be shorter than %(length)i"),
                attribute=name, length=min_length)
        if pattern and not re.match(pattern, value):
            raise exc.InvalidArtifactTypeDefinition(
                message=_("%(attribute)s should match pattern %(pattern)s"),
                attribute=name, pattern=pattern.pattern)


def _build_declarative_meta(cls):
    attrs = dict(cls.__dict__)
    type_name = None
    type_display_name = None
    type_version = None
    type_description = None
    endpoint = None

    for base in cls.__mro__:
        for name, value in six.iteritems(vars(base)):
            if name == '__type_name__':
                if not type_name:
                    type_name = cls.__type_name__
            elif name == '__type_version__':
                if not type_version:
                    type_version = cls.__type_version__
            elif name == '__type_description__':
                if not type_description:
                    type_description = cls.__type_description__
            elif name == '__endpoint__':
                if not endpoint:
                    endpoint = cls.__endpoint__
            elif name == '__type_display_name__':
                if not type_display_name:
                    type_display_name = cls.__type_display_name__
            elif base is not cls and name not in attrs:
                if isinstance(value, AttributeDefinition):
                    attrs[name] = value
                elif isinstance(value, ArtifactPropertyDescriptor):
                    attrs[name] = value.prop

    meta = ArtifactTypeMetadata(type_name=type_name or cls.__name__,
                                type_display_name=type_display_name,
                                type_version=type_version,
                                type_description=type_description,
                                endpoint=endpoint)
    setattr(cls, 'metadata', meta)
    for k, v in attrs.items():
        if k == 'metadata':
            raise exc.InvalidArtifactTypePropertyDefinition(
                _("Cannot declare artifact property with reserved name "
                  "'metadata'"))
        if isinstance(v, AttributeDefinition):
            v._set_name(k)
            wrapper_class = None
            if isinstance(v, ListAttributeDefinition):
                wrapper_class = type("ValidatedList", (list,), {})
                _add_validation_to_list(wrapper_class)
            if isinstance(v, DictAttributeDefinition):
                wrapper_class = type("ValidatedDict", (dict,), {})
                _add_validation_to_dict(wrapper_class)
            prop_descr = ArtifactPropertyDescriptor(v, wrapper_class)
            setattr(cls, k, prop_descr)
            meta.attributes.add(v)


def _validating_method(method, klass):
    def wrapper(self, *args, **kwargs):
        instance_copy = klass(self)
        method(instance_copy, *args, **kwargs)
        self.property.validate(instance_copy)
        method(self, *args, **kwargs)

    return wrapper


def _immutable_method(method):
    def substitution(*args, **kwargs):
        raise exc.InvalidArtifactPropertyValue(
            _("Unable to modify collection in "
              "immutable or readonly property"))

    return substitution


def _add_immutable_wrappers(class_to_add, wrapped_methods):
    for method_name in wrapped_methods:
        method = getattr(class_to_add, method_name, None)
        if method:
            setattr(class_to_add, method_name, _immutable_method(method))


def _add_validation_wrappers(class_to_validate, base_class, validated_methods):
    for method_name in validated_methods:
        method = getattr(class_to_validate, method_name, None)
        if method:
            setattr(class_to_validate, method_name,
                    _validating_method(method, base_class))
    readonly_class = type("Readonly" + class_to_validate.__name__,
                          (class_to_validate,), {})
    _add_immutable_wrappers(readonly_class, validated_methods)

    def __make_immutable__(self):
        return readonly_class(self)

    class_to_validate.__make_immutable__ = __make_immutable__


def _add_validation_to_list(list_based_class):
    validated_methods = ['append', 'extend', 'insert', 'pop', 'remove',
                         'reverse', 'sort', '__setitem__', '__delitem__',
                         '__delslice__']
    _add_validation_wrappers(list_based_class, list, validated_methods)


def _add_validation_to_dict(dict_based_class):
    validated_methods = ['pop', 'popitem', 'setdefault', 'update',
                         '__delitem__', '__setitem__', 'clear']
    _add_validation_wrappers(dict_based_class, dict, validated_methods)


def _kwarg_init_constructor(self, **kwargs):
    self.__suspend_mutability_checks__ = True
    try:
        for k in kwargs:
            if not hasattr(type(self), k):
                raise exc.ArtifactInvalidProperty(prop=k)
            setattr(self, k, kwargs[k])
        self._validate_required(self.metadata.attributes.properties)
    finally:
        del self.__suspend_mutability_checks__


def _validate_required(self, attribute_dict):
    for k, v in six.iteritems(attribute_dict):
        if v.required and (not hasattr(self, k) or getattr(self, k) is None):
            raise exc.InvalidArtifactPropertyValue(name=k, val=None,
                                                   msg=_('Value is required'))


def _update(self, values):
    for k in values:
        if hasattr(type(self), k):
            setattr(self, k, values[k])
        else:
            raise exc.ArtifactInvalidProperty(prop=k)


def _pre_publish_validator(self, *args, **kwargs):
    self._validate_required(self.metadata.attributes.blobs)
    self._validate_required(self.metadata.attributes.dependencies)


_kwarg_init_constructor.__name__ = '__init__'
_pre_publish_validator.__name__ = '__pre_publish__'
_update.__name__ = 'update'


def get_declarative_base(name='base', base_class=object):
    """Returns a base class which should be inherited to construct Artifact
    Type object using the declarative syntax of attribute definition
    """
    bases = not isinstance(base_class, tuple) and (base_class,) or base_class
    class_dict = {'__init__': _kwarg_init_constructor,
                  '_validate_required': _validate_required,
                  '__pre_publish__': _pre_publish_validator,
                  '_declarative_artifact_type': True,
                  'update': _update}
    return ArtifactTypeMetaclass(name, bases, class_dict)
